#!/usr/bin/env python3
import os
import unittest

import numpy as np
import psutil
from pycram.tf_transformations import quaternion_from_euler, quaternion_multiply
from typing_extensions import Optional, List

import pycrap
from pycram.datastructures.dataclasses import ContactPointsList, ContactPoint, AxisAlignedBoundingBox, Color
from pycram.datastructures.enums import Arms, JointType
from pycram.datastructures.pose import PoseStamped
from pycram.robot_description import RobotDescriptionManager
from pycram.world_concepts.world_object import Object
from pycram.validation.error_checkers import calculate_angle_between_quaternions
from pycram.helper import get_robot_description_path, parse_mjcf_actuators
from pycram.object_descriptors.generic import ObjectDescription as GenericObjectDescription

multiverse_installed = True
try:
    from pycram.worlds.multiverse import Multiverse
except ImportError:
    multiverse_installed = False

processes = psutil.process_iter()
process_names = [p.name() for p in processes]
multiverse_running = True
mujoco_running = True
if 'multiverse_server' not in process_names:
    multiverse_running = False
if 'mujoco' not in process_names:
    mujoco_running = False


@unittest.skipIf(not multiverse_installed, "Multiverse is not installed.")
@unittest.skipIf(not multiverse_running, "Multiverse server is not running.")
@unittest.skipIf(not mujoco_running, "Mujoco is not running.")
class MultiversePyCRAMTestCase(unittest.TestCase):
    if multiverse_installed:
        multiverse: Multiverse
    big_bowl: Optional[Object] = None

    @classmethod
    def setUpClass(cls):
        if not multiverse_installed:
            return
        cls.multiverse = Multiverse()

    @classmethod
    def tearDownClass(cls):
        cls.multiverse.exit(remove_saved_states=True)
        cls.multiverse.remove_multiverse_resources()

    def tearDown(self):
        self.multiverse.remove_all_objects()

    def test_load_generic_object(self):
        obj_desc = GenericObjectDescription('test_cube', [0, 0, 0], [0.1, 0.1, 0.1],
                                            color=Color(1, 0, 0, 1))
        obj = Object(obj_desc.name, pycrap.PhysicalObject, description=obj_desc)
        self.assertIsInstance(obj, Object)
        self.assertTrue(obj in self.multiverse.objects)
        obj.set_position([1, 1, 0.1])
        pose = obj.get_pose()
        self.assert_poses_are_equal(pose, PoseStamped.from_list([1, 1, 0.1], [0, 0, 0, 1]))

    def test_save_and_restore_state(self):
        milk = self.spawn_milk([1, 1, 0.1])
        robot = self.spawn_robot()
        apartment = self.spawn_apartment()
        apartment.set_joint_position("cabinet10_drawer1_joint", 0.1)
        robot.attach(milk)
        all_object_attachments = {obj: obj.attachments.copy() for obj in self.multiverse.objects}
        state_id = self.multiverse.save_state()
        robot_link = robot.root_link
        milk_link = milk.root_link
        cid = robot_link.constraint_ids[milk_link]
        self.assertTrue(cid == robot.attachments[milk].id)
        self.multiverse.remove_constraint(cid)
        apartment.set_joint_position("cabinet10_drawer1_joint", 0.0)
        self.multiverse.restore_state(state_id)
        cid = robot_link.constraint_ids[milk_link]
        self.assertTrue(milk_link in robot_link.constraint_ids)
        self.assertTrue(cid == robot.attachments[milk].id)
        for obj in self.multiverse.objects:
            self.assertTrue(len(obj.attachments) == len(all_object_attachments[obj]))
            for att in obj.attachments:
                self.assertTrue(att in all_object_attachments[obj])
        self.assertTrue(apartment.get_joint_position("cabinet10_drawer1_joint") == 0.1)

    def test_spawn_xml_object(self):
        bread = Object("bread_1", pycrap.Bread, "bread_1.xml", pose=PoseStamped.from_list([1, 1, 0.1]))
        self.assert_poses_are_equal(bread.get_pose(), PoseStamped.from_list([1, 1, 0.1]))

    def test_get_axis_aligned_bounding_box_for_one_link_object(self):
        position = [1, 1, 0.1]
        milk = self.spawn_milk(position, quaternion_from_euler(np.pi/4, 0, 0).tolist())
        aabb = milk.get_axis_aligned_bounding_box()
        self.assertIsInstance(aabb, AxisAlignedBoundingBox)
        min_p_1, max_p_1 = aabb.get_min_max()
        width = max_p_1[0] - min_p_1[0]
        length = max_p_1[1] - min_p_1[1]
        height = max_p_1[2] - min_p_1[2]
        self.assertTrue(width > 0)
        self.assertTrue(length > 0)
        self.assertTrue(height > 0)
        # Move the object and check if the bounding box is updated correctly
        position_shift = 1
        milk.set_position([position[0] + position_shift, position[1] + position_shift, 0.1])
        aabb = milk.get_axis_aligned_bounding_box()
        min_p_2, max_p_2 = aabb.get_min_max()
        width_2 = max_p_2[0] - min_p_2[0]
        length_2 = max_p_2[1] - min_p_2[1]
        height_2 = max_p_2[2] - min_p_2[2]
        self.assertAlmostEqual(width, width_2, delta=0.001)
        self.assertAlmostEqual(length, length_2, delta=0.001)
        self.assertAlmostEqual(height, height_2, delta=0.001)
        for i in range(3):
            self.assertAlmostEqual(min_p_1[0] + position_shift, min_p_2[0], delta=0.001)
            self.assertAlmostEqual(max_p_1[0] + position_shift, max_p_2[0], delta=0.001)

    def test_spawn_mesh_object(self):
        milk = Object("milk", pycrap.Milk, "milk.stl", pose=PoseStamped.from_list([1, 1, 0.1]))
        self.assert_poses_are_equal(milk.get_pose(), PoseStamped.from_list([1, 1, 0.1]))
        self.multiverse.simulate(0.2)
        contact_points = milk.contact_points
        self.assertTrue(len(contact_points) > 0)

    def test_parse_mjcf_actuators(self):
        mjcf_file = get_robot_description_path("pal_robotics", "tiago_dual")
        self.assertTrue(os.path.exists(mjcf_file))
        joint_actuators = parse_mjcf_actuators(mjcf_file)
        self.assertIsInstance(joint_actuators, dict)
        self.assertTrue(len(joint_actuators) > 0)
        self.assertTrue("arm_left_1_joint" in joint_actuators)
        self.assertTrue("arm_right_1_joint" in joint_actuators)
        self.assertTrue(joint_actuators["arm_right_1_joint"] == "arm_right_1_actuator")

    def test_get_actuator_for_joint(self):
        robot = self.spawn_robot()
        joint_name = "arm_right_1_joint"
        actuator_name = robot.get_actuator_for_joint(robot.joints[joint_name])
        self.assertEqual(actuator_name, "arm_right_1_actuator")

    def test_get_images_for_target(self):
        robot = self.spawn_robot(robot_name='pr2')
        camera_description = self.multiverse.robot_description.get_default_camera()
        camera_link_name = camera_description.link_name
        camera_pose = robot.get_link_pose(camera_link_name)
        camera_frame = self.multiverse.robot_description.get_camera_frame(robot.name)
        camera_front_facing_axis = camera_description.front_facing_axis
        milk_spawn_position = np.array(camera_front_facing_axis) * 0.5
        orientation = ~(camera_pose.to_transform_stamped(camera_frame)).rotation.to_list()
        milk = self.spawn_milk(milk_spawn_position.tolist(), orientation, frame=camera_frame)
        _, depth, segmentation_mask = self.multiverse.get_images_for_target(milk.pose, camera_pose, plot=False)
        self.assertIsInstance(depth, np.ndarray)
        self.assertIsInstance(segmentation_mask, np.ndarray)
        self.assertTrue(depth.shape == (256, 256))
        self.assertTrue(segmentation_mask.shape == (256, 256))
        self.assertTrue(milk.id in np.unique(segmentation_mask).flatten().tolist())
        avg_depth_of_milk = np.mean(depth[segmentation_mask == milk.id])
        self.assertAlmostEqual(avg_depth_of_milk, 0.5, delta=0.1)

    def test_reset_world(self):
        set_position = [1, 1, 0.1]
        milk = self.spawn_milk(set_position)
        milk.set_position(set_position)
        milk_position = milk.get_position_as_list()
        self.assert_list_is_equal(milk_position[:2], set_position[:2], delta=self.multiverse.conf.position_tolerance)
        self.multiverse.reset_world()
        milk_pose = milk.get_pose()
        self.assert_list_is_equal(milk_pose.position.to_list()[:2],
                                  milk.original_pose.position.to_list()[:2],
                                  delta=self.multiverse.conf.position_tolerance)
        self.assert_orientation_is_equal(milk_pose.orientation.to_list(), milk.original_pose.orientation.to_list()())

    def test_spawn_robot_with_actuators_directly_from_multiverse(self):
        if self.multiverse.conf.use_controller:
            robot_name = "tiago_dual"
            rdm = RobotDescriptionManager()
            rdm.load_description(robot_name)
            self.multiverse.spawn_robot_with_controller(robot_name, PoseStamped())

    def test_spawn_object(self):
        milk = self.spawn_milk([1, 1, 0.1])
        self.assertIsInstance(milk, Object)
        milk_pose = milk.get_pose()
        self.assert_list_is_equal(milk_pose.position.to_list()[:2], [1, 1],
                                  delta=self.multiverse.conf.position_tolerance)
        self.assert_orientation_is_equal(milk_pose.orientation.to_list(), milk.original_pose.orientation.to_list()())

    def test_remove_object(self):
        milk = self.spawn_milk([1, 1, 0.1])
        milk.remove()
        self.assertTrue(milk not in self.multiverse.objects)
        self.assertFalse(self.multiverse.check_object_exists(milk))

    def test_check_object_exists(self):
        milk = self.spawn_milk([1, 1, 0.1])
        self.assertTrue(self.multiverse.check_object_exists(milk))

    def test_set_position(self):
        milk = self.spawn_milk([1, 1, 0.1])
        original_milk_position = milk.get_position_as_list()
        original_milk_position[0] += 1
        milk.set_position(original_milk_position)
        milk_position = milk.get_position_as_list()
        self.assert_list_is_equal(milk_position[:2], original_milk_position[:2],
                                  delta=self.multiverse.conf.position_tolerance)

    def test_update_position(self):
        milk = self.spawn_milk([1, 1, 0.1])
        milk_position = milk.get_position_as_list()
        self.assert_list_is_equal(milk_position[:2], [1, 1], delta=self.multiverse.conf.position_tolerance)

    def test_set_joint_position(self):
        if self.multiverse.robot is None:
            robot = self.spawn_robot()
        else:
            robot = self.multiverse.robot
        step = 0.2
        for joint in ['torso_lift_joint']:
            joint_type = robot.joints[joint].type
            original_joint_position = robot.get_joint_position(joint)
            robot.set_joint_position(joint, original_joint_position + step)
            joint_position = robot.get_joint_position(joint)
            if not self.multiverse.conf.use_controller:
                delta = self.multiverse.conf.prismatic_joint_position_tolerance if joint_type == JointType.PRISMATIC \
                    else self.multiverse.conf.revolute_joint_position_tolerance
            else:
                delta = 0.18
            self.assertAlmostEqual(joint_position, original_joint_position + step, delta=delta)

    def test_spawn_robot(self):
        if self.multiverse.robot is not None:
            robot = self.multiverse.robot
        else:
            robot = self.spawn_robot(robot_name="pr2")
        self.assertIsInstance(robot, Object)
        self.assertTrue(robot in self.multiverse.objects)
        self.assertTrue(self.multiverse.robot.name == robot.name)

    def test_destroy_robot(self):
        if self.multiverse.robot is None:
            self.spawn_robot()
        self.assertTrue(self.multiverse.robot in self.multiverse.objects)
        self.multiverse.robot.remove()
        self.assertTrue(self.multiverse.robot not in self.multiverse.objects)

    def test_respawn_robot(self):
        self.spawn_robot()
        self.assertTrue(self.multiverse.robot in self.multiverse.objects)
        self.multiverse.robot.remove()
        self.assertTrue(self.multiverse.robot not in self.multiverse.objects)
        self.spawn_robot()
        self.assertTrue(self.multiverse.robot in self.multiverse.objects)

    def test_set_robot_position(self):
        step = -1
        for i in range(3):
            self.spawn_robot()
            new_position = [-3 + step * i, -3 + step * i, 0.001]
            self.multiverse.robot.set_position(new_position)
            robot_position = self.multiverse.robot.get_position_as_list()
            self.assert_list_is_equal(robot_position[:2], new_position[:2],
                                      delta=self.multiverse.conf.position_tolerance)
            self.tearDown()

    def test_set_robot_orientation(self):
        self.spawn_robot()
        for i in range(3):
            current_quaternion = self.multiverse.robot.get_orientation_as_list()
            # rotate by 45 degrees without using euler angles
            rotation_quaternion = quaternion_from_euler(0, 0, np.pi / 4)
            new_quaternion = quaternion_multiply(current_quaternion, rotation_quaternion)
            self.multiverse.robot.set_orientation(new_quaternion)
            robot_orientation = self.multiverse.robot.get_orientation_as_list()
            quaternion_difference = calculate_angle_between_quaternions(new_quaternion, robot_orientation)
            self.assertAlmostEqual(quaternion_difference, 0, delta=self.multiverse.conf.orientation_tolerance)

    def test_set_robot_pose(self):
        self.spawn_robot(orientation=quaternion_from_euler(0, 0, np.pi / 4))
        position_step = -1
        angle_step = np.pi / 4
        num_steps = 10
        self.step_robot_pose(self.multiverse.robot, position_step, angle_step, num_steps)
        position_step = 1
        angle_step = -np.pi / 4
        self.step_robot_pose(self.multiverse.robot, position_step, angle_step, num_steps)

    def step_robot_pose(self, robot, position_step, angle_step, num_steps):
        original_position = robot.get_position_as_list()
        original_orientation = robot.get_orientation_as_list()
        for i in range(num_steps):
            new_position = [original_position[0] + position_step * (i + 1),
                            original_position[1] + position_step * (i + 1), original_position[2]]
            rotation_quaternion = quaternion_from_euler(0, 0, angle_step * (i + 1))
            new_quaternion = quaternion_multiply(original_orientation, rotation_quaternion)
            new_pose = PoseStamped(new_position, new_quaternion)
            self.multiverse.robot.set_pose(new_pose)
            robot_pose = self.multiverse.robot.get_pose()
            self.assert_poses_are_equal(new_pose, robot_pose,
                                        position_delta=self.multiverse.conf.position_tolerance,
                                        orientation_delta=self.multiverse.conf.orientation_tolerance)

    def test_get_environment_pose(self):
        if "apartment" not in self.multiverse.get_object_names():
            apartment = Object("apartment", pycrap.Apartment, f"apartment.urdf")
        else:
            apartment = self.multiverse.get_object_by_name("apartment")
        pose = apartment.get_pose()
        self.assertIsInstance(pose, PoseStamped)

    def test_attach_object(self):
        for _ in range(3):
            milk = self.spawn_milk([1, 0.1, 0.1])
            cup = self.spawn_cup([1, 1.1, 0.1])
            milk.attach(cup)
            self.assertTrue(cup in milk.attachments)
            milk_position = milk.get_position_as_list()
            milk_position[0] += 1
            cup_position = cup.get_position_as_list()
            estimated_cup_position = cup_position.copy()
            estimated_cup_position[0] += 1
            milk.set_position(milk_position)
            new_cup_position = cup.get_position_as_list()
            self.assert_list_is_equal(new_cup_position[:2], estimated_cup_position[:2],
                                      self.multiverse.conf.position_tolerance)
            self.tearDown()

    def test_detach_object(self):
        for i in range(2):
            milk = self.spawn_milk([1, 0, 0.1])
            cup = self.spawn_cup([1, 1, 0.1])
            milk.attach(cup)
            self.assertTrue(cup in milk.attachments)
            milk.detach(cup)
            self.assertTrue(cup not in milk.attachments)
            milk_position = milk.get_position_as_list()
            milk_position[0] += 1
            cup_position = cup.get_position_as_list()
            estimated_cup_position = cup_position.copy()
            milk.set_position(milk_position)
            new_milk_position = milk.get_position_as_list()
            new_cup_position = cup.get_position_as_list()
            self.assert_list_is_equal(new_milk_position[:2], milk_position[:2],
                                      self.multiverse.conf.position_tolerance)
            self.assert_list_is_equal(new_cup_position[:2], estimated_cup_position[:2],
                                      self.multiverse.conf.position_tolerance)
            self.tearDown()

    def test_attach_with_robot(self):
        milk = self.spawn_milk([-1, -1, 0.1])
        robot = self.spawn_robot()
        ee_link = self.multiverse.get_arm_tool_frame_link(Arms.RIGHT)
        # Get position of milk relative to robot end effector
        robot.attach(milk, ee_link.name)
        self.assertTrue(robot in milk.attachments)
        milk_initial_pose = milk.root_link.get_pose_wrt_link(ee_link)
        robot_position = 1.57
        robot.set_joint_position("arm_right_2_joint", robot_position)
        milk_pose = milk.root_link.get_pose_wrt_link(ee_link)
        self.assert_poses_are_equal(milk_initial_pose, milk_pose)

    def test_get_object_contact_points(self):
        for i in range(3):
            milk = self.spawn_milk([1, 1, 0.01], [0, -0.707, 0, 0.707])
            contact_points = self.multiverse.get_object_contact_points(milk)
            self.assertIsInstance(contact_points, ContactPointsList)
            self.assertTrue(len(contact_points) >= 1)
            self.assertIsInstance(contact_points[0], ContactPoint)
            self.assertTrue(contact_points[0].body_b.object, self.multiverse.floor)
            cup = self.spawn_cup([1, 1, 0.12])
            # This is needed because the cup is spawned in the air, so it needs to fall
            # to get in contact with the milk
            self.multiverse.simulate(0.4)
            contact_points = self.multiverse.get_object_contact_points(cup)
            self.assertIsInstance(contact_points, ContactPointsList)
            self.assertTrue(len(contact_points) >= 1)
            self.assertIsInstance(contact_points[0], ContactPoint)
            self.assertTrue(contact_points[0].body_b.object, milk)
            self.tearDown()

    def test_get_robot_contact_points(self):
        robot = self.spawn_robot([0.9345829872370865, 1.9027591011850133, 0.0],
                                 quaternion_from_euler(0, 0, 2.26).tolist(),
                                 robot_name="pr2")
        apartment = self.spawn_apartment()
        contact_points = self.multiverse.get_contact_points_between_two_bodies(robot, apartment)
        self.assertTrue(len(contact_points) > 0)

    def test_get_contact_points_between_two_objects(self):
        for i in range(3):
            milk = self.spawn_milk([1, 1, 0.01], [0, -0.707, 0, 0.707])
            cup = self.spawn_cup([1, 1, 0.12])
            # This is needed because the cup is spawned in the air so it needs to fall
            # to get in contact with the milk
            self.multiverse.simulate(0.4)
            contact_points = self.multiverse.get_contact_points_between_two_bodies(milk, cup)
            self.assertIsInstance(contact_points, ContactPointsList)
            self.assertTrue(len(contact_points) >= 1)
            self.assertIsInstance(contact_points[0], ContactPoint)
            self.assertTrue(contact_points[0].body_a.object, milk)
            self.assertTrue(contact_points[0].body_b.object, cup)
            self.tearDown()

    def test_get_one_ray(self):
        milk = self.spawn_milk([1, 1, 0.1])
        intersected_object = self.multiverse._ray_test([1, 2, 0.1], [1, 1.5, 0.1])
        self.assertTrue(intersected_object is None)
        intersected_object = self.multiverse._ray_test([1, 2, 0.1], [1, 1, 0.1])
        self.assertTrue(intersected_object == milk.id)

    def test_get_rays(self):
        milk = self.spawn_milk([1, 1, 0.1])
        intersected_objects = self.multiverse._ray_test_batch([[1, 2, 0.1], [1, 2, 0.1]],
                                                              [[1, 1.5, 0.1], [1, 1, 0.1]])
        self.assertTrue(intersected_objects[0][0] == -1)
        self.assertTrue(intersected_objects[1][0] == milk.id)

    @staticmethod
    def spawn_big_bowl() -> Object:
        big_bowl = Object("big_bowl", pycrap.Bowl, "BigBowl.obj",
                          pose=PoseStamped.from_list([2, 2, 0.1], [0, 0, 0, 1]))
        return big_bowl

    @staticmethod
    def spawn_milk(position: List, orientation: Optional[List] = None, frame="map") -> Object:
        if orientation is None:
            orientation = [0, 0, 0, 1]
        milk = Object("milk_box", pycrap.Milk, "milk_box.xml",
                      pose=PoseStamped(position, orientation, frame_id=frame))
        return milk

    def spawn_apartment(self) -> Object:
        if "apartment" not in self.multiverse.get_object_names():
            apartment = Object("apartment", pycrap.Apartment, f"apartment.urdf")
        else:
            apartment = self.multiverse.get_object_by_name("apartment")
        return apartment

    def spawn_robot(self, position: Optional[List[float]] = None,
                    orientation: Optional[List[float]] = None,
                    robot_name: Optional[str] = 'tiago_dual',
                    replace: Optional[bool] = True) -> Object:
        if position is None:
            position = [-2, -2, 0.001]
        if orientation is None:
            orientation = [0, 0, 0, 1]
        if self.multiverse.robot is None or replace:
            if self.multiverse.robot is not None:
                self.multiverse.robot.remove()
            robot = Object(robot_name, pycrap.Robot, f"{robot_name}.urdf",
                           pose=PoseStamped(position, orientation))
        else:
            robot = self.multiverse.robot
            robot.set_position(position)
        return robot

    @staticmethod
    def spawn_cup(position: List) -> Object:
        cup = Object("cup", pycrap.Cup, "cup.xml",
                     pose=PoseStamped(position, [0, 0, 0, 1]))
        return cup

    def assert_poses_are_equal(self, pose1: PoseStamped, pose2: PoseStamped,
                               position_delta: Optional[float] = None, orientation_delta: Optional[float] = None):
        if position_delta is None:
            position_delta = self.multiverse.conf.position_tolerance
        if orientation_delta is None:
            orientation_delta = self.multiverse.conf.orientation_tolerance
        self.assert_position_is_equal(pose1.position.to_list(), pose2.position.to_list(), delta=position_delta)
        self.assert_orientation_is_equal(pose1.orientation.to_list(), pose2.orientation.to_list(),
                                         delta=orientation_delta)

    def assert_position_is_equal(self, position1: List[float], position2: List[float], delta: Optional[float] = None):
        if delta is None:
            delta = self.multiverse.conf.position_tolerance
        self.assert_list_is_equal(position1, position2, delta=delta)

    def assert_orientation_is_equal(self, orientation1: List[float], orientation2: List[float],
                                    delta: Optional[float] = None):
        if delta is None:
            delta = self.multiverse.conf.orientation_tolerance
        self.assertAlmostEqual(calculate_angle_between_quaternions(orientation1, orientation2), 0, delta=delta)

    def assert_list_is_equal(self, list1: List, list2: List, delta: float):
        for i in range(len(list1)):
            self.assertAlmostEqual(list1[i], list2[i], delta=delta)



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import psutil
import random_events
from matplotlib import colors
from probabilistic_model.probabilistic_circuit.nx.helper import uniform_measure_of_event
from probabilistic_model.probabilistic_circuit.nx.probabilistic_circuit import ProbabilisticCircuit
from random_events.interval import Interval, reals, closed_open, closed
from random_events.product_algebra import Event, SimpleEvent
from random_events.variable import Continuous
from .tf_transformations import quaternion_from_matrix, quaternion_from_euler
from typing_extensions import Tuple, List, Optional, Iterator

from .datastructures.dataclasses import BoxVisualShape, Color
from .datastructures.pose import TransformStamped
from .ros import logwarn
from .datastructures.dataclasses import AxisAlignedBoundingBox
from .datastructures.pose import PoseStamped
from .datastructures.world import UseProspectionWorld
from .datastructures.world import World
from .description import Link
from .local_transformer import LocalTransformer
from .ros import wait_for_message
from .utils import chunks
from .world_concepts.world_object import Object

try:
    from nav_msgs.msg import OccupancyGrid, MapMetaData
except ImportError:
    pass


@dataclass
class Rectangle:
    """
    A rectangle that is described by a lower and upper x and y value.
    """
    x_lower: float
    x_upper: float
    y_lower: float
    y_upper: float

    def translate(self, x: float, y: float):
        """Translate the rectangle by x and y"""
        self.x_lower += x
        self.x_upper += x
        self.y_lower += y
        self.y_upper += y

    def scale(self, x_factor: float, y_factor: float):
        """Scale the rectangle by x_factor and y_factor"""
        self.x_lower *= x_factor
        self.x_upper *= x_factor
        self.y_lower *= y_factor
        self.y_upper *= y_factor


class Costmap:
    """
    The base class of all Costmaps which implements the visualization of costmaps
    in the World.
    """

    def __init__(self, resolution: float,
                 height: int,
                 width: int,
                 origin: PoseStamped,
                 map: np.ndarray,
                 world: Optional[World] = None):
        """
        The constructor of the base class of all Costmaps.

        :param resolution: The distance in metre in the real-world which is
         represented by a single entry in the costmap.
        :param height: The height of the costmap.
        :param width: The width of the costmap.
        :param origin: The origin of the costmap, in world coordinate frame. The origin of the costmap is located in the
         centre of the costmap.
        :param map: The costmap represents as a 2D numpy array.
        :param world: The World for which the costmap should be created.
        """
        self.world = world if world else World.current_world
        self.resolution: float = resolution
        self.size: int = height
        self.height: int = height
        self.width: int = width
        local_transformer = LocalTransformer()
        self.origin: PoseStamped = local_transformer.transform_pose(origin, 'map')
        self.map: np.ndarray = map
        self.vis_ids: List[int] = []

    def visualize(self) -> None:
        """
        Visualizes a costmap in the BulletWorld, the visualisation works by
        subdividing the costmap in rectangles which are then visualized as pybullet
        visual shapes.
        """
        if self.vis_ids != []:
            return

        # working on a copy of the costmap, since found rectangles are deleted
        map = np.copy(self.map)
        boxes = []
        # Finding all rectangles in the costmap
        for i in range(0, map.shape[0]):
            for j in range(0, map.shape[1]):
                if map[i][j] > 0:
                    curr_width = self._find_consectuive_line((i, j), map)
                    curr_pose = (i, j)
                    curr_height = self._find_max_box_height((i, j), curr_width, map)
                    avg = np.average(map[i:i + curr_height, j:j + curr_width])
                    boxes.append([curr_pose, curr_height, curr_width, avg])
                    map[i:i + curr_height, j:j + curr_width] = 0
        cells = []
        # Creation of the visual shapes, for documentation of the visual shapes
        # please look here: https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit#heading=h.q1gn7v6o58bf
        for box in boxes:
            box = BoxVisualShape(Color(1, 0, 0, 0.6),
                                 [(box[0][0] + box[1] / 2) * self.resolution,
                                  (box[0][1] + box[2] / 2) * self.resolution, 0.],
                                 [(box[1] * self.resolution) / 2, (box[2] * self.resolution) / 2, 0.001])
            visual = self.world.create_visual_shape(box)
            cells.append(visual)
        # Set to 127 for since this is the maximal amount of links in a multibody
        for cell_parts in self._chunks(cells, 127):
            offset = [[-self.height / 2 * self.resolution, -self.width / 2 * self.resolution, 0.05], [0, 0, 0, 1]]
            origin_transform = TransformStamped.from_list(self.origin.position.to_list(),
                                                          self.origin.orientation.to_list())
            offset_transform = TransformStamped.from_list(offset[0], offset[1])
            new_pose_transform = origin_transform * offset_transform
            new_pose = PoseStamped.from_list(new_pose_transform.translation.to_list(),
                                             new_pose_transform.rotation.to_list())
            map_obj = self.world.create_multi_body_from_visual_shapes(cell_parts, new_pose)
            self.vis_ids.append(map_obj)

    def _chunks(self, lst: List, n: int) -> List:
        """
        Yield successive n-sized chunks from lst.

        :param lst: The list from which chunks should be yielded
        :param n: Size of the chunks
        :return: A list of size n from lst
        """
        for i in range(0, len(lst), n):
            yield lst[i:i + n]

    def close_visualization(self) -> None:
        """
        Removes the visualization from the World.
        """
        for v_id in self.vis_ids:
            self.world.remove_visual_object(v_id)
        self.vis_ids = []

    def _find_consectuive_line(self, start: Tuple[int, int], map: np.ndarray) -> int:
        """
        Finds the number of consecutive entries in the costmap which are greater
        than zero.

        :param start: The indices in the costmap from which the consecutive line should be found.
        :param map: The costmap in which the line should be found.
        :return: The length of the consecutive line of entries greater than zero.
        """
        width = map.shape[1]
        lenght = 0
        for i in range(start[1], width):
            if map[start[0]][i] > 0:
                lenght += 1
            else:
                return lenght
        return lenght

    def _find_max_box_height(self, start: Tuple[int, int], length: int, map: np.ndarray) -> int:
        """
        Finds the maximal height for a rectangle with a given width in a costmap.
        The method traverses one row at a time and checks if all entries for the
        given width are greater than zero. If an entry is less or equal than zero
        the height is returned.

        :param start: The indices in the costmap from which the method should start.
        :param length: The given width for the rectangle
        :param map: The costmap in which should be searched.
        :return: The height of the rectangle.
        """
        height, width = map.shape
        curr_height = 1
        for i in range(start[0], height):
            for j in range(start[1], start[1] + length):
                if map[i][j] <= 0:
                    return curr_height
            curr_height += 1
        return curr_height

    def merge(self, other_cm: Costmap) -> Costmap:
        """
        Merges the values of two costmaps and returns a new costmap that has for
        every cell the merged values of both inputs. To merge two costmaps they
        need to fulfill 3 constrains:

        1. They need to have the same size
        2. They need to have the same x and y coordinates in the origin
        3. They need to have the same resolution

        If any of these constrains is not fulfilled a ValueError will be raised.

        :param other_cm: The other costmap with which this costmap should be merged.
        :return: A new costmap that contains the merged values
        """
        if self.size != other_cm.size:
            raise ValueError("You can only merge costmaps of the same size.")
        elif self.origin.position.x != other_cm.origin.position.x or self.origin.position.y != other_cm.origin.position.y \
                or self.origin.orientation != other_cm.origin.orientation:
            raise ValueError("To merge costmaps, the x and y coordinate as well as the orientation must be equal.")
        elif self.resolution != other_cm.resolution:
            raise ValueError("To merge two costmaps their resolution must be equal.")
        new_map = np.zeros((self.height, self.width))
        # A numpy array of the positions where both costmaps are greater than 0
        merge = np.logical_and(self.map > 0, other_cm.map > 0)
        new_map[merge] = self.map[merge] * other_cm.map[merge]
        max_val = np.max(new_map)
        if max_val != 0:
            new_map = (new_map / np.max(new_map)).reshape((self.height, self.width))
        else:
            new_map = new_map.reshape((self.height, self.width))
            logwarn("Merged costmap is empty.")
        return Costmap(self.resolution, self.height, self.width, self.origin, new_map)

    def __add__(self, other: Costmap) -> Costmap:
        """
        Overloading of the "+" operator for merging of Costmaps. Furthermore, checks if 'other' is actual a Costmap and
        raises a ValueError if this is not the case. Please check :func:`~Costmap.merge` for further information of merging.

        :param other: Another Costmap
        :return: A new Costmap that contains the merged values from this Costmap and the other Costmap
        """
        if isinstance(other, Costmap):
            return self.merge(other)
        else:
            raise ValueError(f"Can only combine two costmaps other type was {type(other)}")

    def partitioning_rectangles(self) -> List[Rectangle]:
        """
        Partition the map attached to this costmap into rectangles. The rectangles are axis aligned, exhaustive and
        disjoint sets.

        :return: A list containing the partitioning rectangles
        """
        ocm_map = np.copy(self.map)
        origin = np.array([self.height / 2, self.width / 2]) * -1
        rectangles = []

        # for every index pair (i, j) in the occupancy costmap
        for i in range(0, self.map.shape[0]):
            for j in range(0, self.map.shape[1]):

                # if this index has not been used yet
                if ocm_map[i][j] > 0:
                    curr_width = self._find_consectuive_line((i, j), ocm_map)
                    curr_pose = (i, j)
                    curr_height = self._find_max_box_height((i, j), curr_width, ocm_map)

                    # calculate the rectangle in the costmap
                    x_lower = curr_pose[0]
                    x_upper = curr_pose[0] + curr_height
                    y_lower = curr_pose[1]
                    y_upper = curr_pose[1] + curr_width

                    # mark the found rectangle as occupied
                    ocm_map[i:i + curr_height, j:j + curr_width] = 0

                    # transform rectangle to map space
                    rectangle = Rectangle(x_lower, x_upper, y_lower, y_upper)
                    rectangle.translate(*origin)
                    rectangle.scale(self.resolution, self.resolution)
                    rectangles.append(rectangle)

        return rectangles


class OccupancyCostmap(Costmap):
    """
    The occupancy Costmap represents a map of the environment where obstacles or
    positions which are inaccessible for a robot have a value of -1.
    """

    def __init__(self, distance_to_obstacle: float,
                 from_ros: Optional[bool] = False,
                 size: Optional[int] = 100,
                 resolution: Optional[float] = 0.02,
                 origin: Optional[PoseStamped] = None,
                 world: Optional[World] = None):
        """
        Constructor for the Occupancy costmap, the actual costmap is received
        from the ROS map_server and wrapped by this class. Meta-data about the
        costmap is also received from the map_server.

        :param distance_to_obstacle: The distance by which the obstacles should be
            inflated. Meaning that obstacles in the costmap are growing bigger by this
            distance.
        :param from_ros: This determines if the Occupancy map should be created
            from the map provided by the ROS map_server or from the World.
            If True then the map from the ROS map_server will be used otherwise
            the Occupancy map will be created from the World.
        :param size: The length of the side of the costmap. The costmap will be created
            as a square. This will only be used if from_ros is False.
        :param resolution: The resolution of this costmap. This determines how much
            meter one pixel in the costmap represents. This is only used if from_ros
            is False.
        :param origin: This determines the origin of the costmap. The origin will
            be in the middle of the costmap. This parameter is only used if from_ros
            is False.
        """
        self.world = world if world else World.current_world
        if from_ros:
            meta = self._get_map_metadata()
            self.original_map = np.reshape(self._get_map(), (meta.height, meta.width))
            self.meta_origin = [meta.origin.position.x, meta.origin.position.y, meta.origin.position.z]
            self.resolution = meta.resolution
            self.height = meta.height
            self.width = meta.width
            # Nunber of cells that have to be between a valid cell and an obstacle
            self.distance_obstacle = max(int(distance_to_obstacle / self.resolution), 1)
            Costmap.__init__(self, meta.resolution, meta.height, meta.width,
                             self._calculate_diff_origin(meta.height, meta.width),
                             np.rot90(np.flip(self._convert_map(self.original_map), 0)))
        else:
            self.size = size
            self.origin = PoseStamped.from_list() if not origin else origin
            lt = LocalTransformer()
            self.origin = lt.transform_pose(self.origin, "map")
            self.resolution = resolution
            self.distance_obstacle = max(int(distance_to_obstacle / self.resolution), 1)
            self.map = self._create_from_world(size, resolution)
            Costmap.__init__(self, resolution, size, size, self.origin, self.map)

    def _calculate_diff_origin(self, height: int, width: int) -> PoseStamped:
        """
        Calculates the difference between the origin of the costmap
        as stated by the meta-data and the actual middle of the costmap which
        is used by PyCRAM to visualize the costmap. The origin as stated by the
        meta-data refers to the position of the global coordinate frame with
        the bottom left corner as reference.

        :param height: The height of the costmap
        :param width: The width of the costmap
        :return: The difference between the actual origin and center of the costmap
        """
        actual_origin = [int(height / 2) * self.resolution, int(width / 2) * self.resolution, 0]
        origin = np.array(self.meta_origin) + np.array(actual_origin)
        return PoseStamped.from_list(origin.tolist())

    @staticmethod
    def _get_map() -> np.ndarray:
        """
        Receives the map array from the map_server converts it and into a numpy array.

        :return: The costmap as a numpy array.
        """
        print("Waiting for Map")
        map = wait_for_message("/map", OccupancyGrid)
        print("Recived Map")
        return np.array(map.data)

    @staticmethod
    def _get_map_metadata() -> MapMetaData:
        """
        Receives the meta-data about the costmap from the map_server and returns it.
        The meta-data contains things like, height, width, origin and resolution.

        :return: The meta-data for the costmap array.
        """
        print("Waiting for Map Meta Data")
        meta = wait_for_message("/map_metadata", MapMetaData)
        print("Recived Meta Data")
        return meta

    def _convert_map(self, map: np.ndarray) -> np.ndarray:
        """
        Converts the Occupancy Map received from ROS to be more consistent
        with how PyCRAM handles its costmap. Every possible cell for a robot to stand
        is set to one while anything else is set to zero. Additionally, this method
        also takes into account the distance_to_obstacle parameter and sets cell values
        that are too close to an obstacle to 0.

        :param map: The map that should be converted. Represented as 2d numpy array
        :return: The converted map. Represented as 2d numpy array.
        """
        map = np.pad(map, (int(self.distance_obstacle / 2), int(self.distance_obstacle / 2)))

        sub_shape = (self.distance_obstacle, self.distance_obstacle)
        view_shape = tuple(np.subtract(map.shape, sub_shape) + 1) + sub_shape
        strides = map.strides + map.strides

        sub_matrices = np.lib.stride_tricks.as_strided(map, view_shape, strides)
        sub_matrices = sub_matrices.reshape(sub_matrices.shape[:-2] + (-1,))
        sum = np.sum(sub_matrices, axis=2)
        return (sum == 0).astype('int16')

    def create_sub_map(self, sub_origin: PoseStamped, size: int) -> Costmap:
        """
        Creates a smaller map from the overall occupancy map, the new map is centered
        around the point specified by "sub_origin" and has the size "size". The
        resolution of the costmap stays the same for the sub costmap.

        :param sub_origin: The point in global coordinate frame, around which the sub costmap should be centered.
        :param size: The size the sub costmap should have.
        :return: The sub costmap, represented as 2d numpy array.
        """
        # To ensure this is a numpy array
        sub_origin = np.array(sub_origin.position.to_list())
        # Since origin obtained from the meta data uses bottom left corner as reference.
        sub_origin *= -1
        # Calculates origin of sub costmap as vector between origin and given sub_origin
        new_origin = np.array(self.meta_origin) + sub_origin
        # Convert from vector in meter to index values
        new_origin /= self.resolution
        new_origin = np.abs(new_origin)
        # Offset to top left corner, for easier slicing
        new_origin = (new_origin - size / 2).astype(int)

        # slices a submap with size "size" around the given origin
        sub_map = self.original_map[new_origin[1]: new_origin[1] + size,
                  new_origin[0]: new_origin[0] + size]
        # Convert map to fit with the other costmaps
        sub_map = np.rot90(np.flip(self._convert_map(sub_map), 0))
        return Costmap(self.resolution, size, size, PoseStamped.from_list(list(sub_origin * -1)), sub_map)

    def _create_from_world(self, size: int, resolution: float) -> np.ndarray:
        """
        Creates an Occupancy Costmap for the specified World.
        This map marks every position as valid that has no object above it. After
        creating the costmap the distance to obstacle parameter is applied.

        :param size: The size of this costmap. The size specifies the length of one side of the costmap. The costmap is created as a square.
        :param resolution: The resolution of this costmap. This determines how much meter a pixel in the costmap represents.
        """
        origin_position = self.origin.position.to_list()
        # Generate 2d grid with indices
        indices = np.concatenate(np.dstack(np.mgrid[int(-size / 2):int(size / 2), int(-size / 2):int(size / 2)]),
                                 axis=0) * resolution + np.array(origin_position[:2])
        # Add the z-coordinate to the grid, which is either 0 or 10
        indices_0 = np.pad(indices, (0, 1), mode='constant', constant_values=5)[:-1]
        indices_10 = np.pad(indices, (0, 1), mode='constant', constant_values=0)[:-1]
        # Zips both arrays such that there are tuples for every coordinate that
        # only differ in the z-coordinate
        rays = np.dstack(np.dstack((indices_0, indices_10))).T

        res = np.zeros(len(rays))
        # Using the World rayTest to check if there is an object above the position
        # if there is no object the position is marked as valid
        # 16383 is the maximal number of rays that can be processed in a batch
        i = 0
        j = 0
        floor_id = self.world.get_object_by_name("floor").id
        for n in chunks(np.array(rays), World.current_world.conf.max_batch_size_for_rays):
            r_t = World.current_world.ray_test_batch(n[:, 0], n[:, 1], num_threads=0)
            while r_t is None:
                r_t = World.current_world.ray_test_batch(n[:, 0], n[:, 1], num_threads=0)
            j += len(n)
            if World.robot:
                attached_objs_id = [o.id for o in self.world.robot.attachments.keys()]
                res[i:j] = [
                    1 if ray.obj_id in [-1, self.world.robot.id, floor_id] + attached_objs_id else 0 for
                    ray in r_t]
            else:
                res[i:j] = [1 if ray.obj_id in [-1, floor_id] else 0 for ray in r_t]
            i += len(n)

        res = np.flip(np.reshape(np.array(res), (size, size)))

        map = np.pad(res, (int(self.distance_obstacle / 2), int(self.distance_obstacle / 2)))

        sub_shape = (self.distance_obstacle * 2, self.distance_obstacle * 2)
        view_shape = tuple(np.subtract(map.shape, sub_shape) + 1) + sub_shape
        strides = map.strides + map.strides

        sub_matrices = np.lib.stride_tricks.as_strided(map, view_shape, strides)
        sub_matrices = sub_matrices.reshape(sub_matrices.shape[:-2] + (-1,))

        sum = np.sum(sub_matrices, axis=2)
        map = (sum == (self.distance_obstacle * 2) ** 2).astype('int16')
        # The map loses some size due to the strides and because I dont want to
        # deal with indices outside of the index range
        offset = self.size - map.shape[0]
        odd = 0 if offset % 2 == 0 else 1
        map = np.pad(map, (offset // 2, offset // 2 + odd))

        return np.flip(map)


class VisibilityCostmap(Costmap):
    """
    A costmap that represents the visibility of a specific point for every position around
    this point. For a detailed explanation on how the creation of the costmap works
    please look here: `PhD Thesis (page 173) <https://mediatum.ub.tum.de/doc/1239461/1239461.pdf>`_
    """

    def __init__(self, min_height: float,
                 max_height: float,
                 size: Optional[int] = 100,
                 resolution: Optional[float] = 0.02,
                 origin: Optional[PoseStamped] = None,
                 world: Optional[World] = None,
                 target_object: Optional[Object] = None,
                 robot: Optional[Object] = None):
        """
        Visibility Costmaps show for every position around the origin pose if the origin can be seen from this pose.
        The costmap is able to deal with height differences of the camera while in a single position, for example, if
        the robot has a movable torso.

        :param min_height: This is the minimal height the camera can be. This parameter
            is mostly relevant if the vertical position of the camera can change.
        :param max_height: This is the maximal height the camera can be. This is
            mostly relevant if the vertical position of the camera can change.
        :param size: The length of the side of the costmap, the costmap is created
            as a square.
        :param resolution: This parameter specifies how much meter a pixel in the
            costmap represents.
        :param origin: The pose in world coordinate frame around which the
            costmap should be created.
        :param world: The World for which the costmap should be created.
        :param target_object: The object that should be visible.
        :param robot: The robot for which the visibility costmap should be created.
        """
        if (11 * size ** 2 + size ** 3) * 2 > psutil.virtual_memory().available:
            raise OSError("Not enough free RAM to calculate a costmap of this size")

        self.world = world if world else World.current_world
        self.map = np.zeros((size, size))
        self.size = size
        self.resolution = resolution
        # for pr2 = 1.27
        self.max_height: float = max_height
        # for pr2 = 1.6
        self.min_height: float = min_height
        self.origin: PoseStamped = PoseStamped.from_list() if not origin else origin
        self.target_object: Optional[Object] = target_object
        self.robot: Optional[Object] = robot
        self._generate_map()
        Costmap.__init__(self, resolution, size, size, self.origin, self.map)

    @property
    def robot(self) -> Optional[Object]:
        return self._robot

    @robot.setter
    def robot(self, robot: Optional[Object]) -> None:
        if robot is not None:
            self._robot = World.current_world.get_prospection_object_for_object(robot)
            self.robot_original_pose = self._robot.pose
        else:
            self._robot = None
            self.robot_original_pose = None

    @property
    def target_object(self) -> Optional[Object]:
        return self._target_object

    @target_object.setter
    def target_object(self, target_object: Optional[Object]) -> None:
        if target_object is not None and not isinstance(target_object, PoseStamped):
            self._target_object = World.current_world.get_prospection_object_for_object(target_object)
            self.target_original_pose = self._target_object.pose
        else:
            self._target_object = None
            self.target_original_pose = None

    def move_target_and_robot_far_away(self):
        if self.target_object is not None:
            self.target_object.set_pose(PoseStamped.from_list([self.origin.position.x + self.size * self.resolution * 2,
                                                               self.origin.position.y + self.size * self.resolution * 2,
                                                               self.target_original_pose.position.z]))
        if self.robot is not None:
            self.robot.set_pose(PoseStamped.from_list([self.origin.position.x + self.size * self.resolution * 3,
                                                       self.origin.position.y + self.size * self.resolution * 3,
                                                       self.robot_original_pose.position.z]))

    def return_target_and_robot_to_their_original_position(self):
        if self.target_original_pose is not None:
            self.target_object.set_pose(self.target_original_pose)
        if self.robot_original_pose is not None:
            self.robot.set_pose(self.robot_original_pose)

    def _create_images(self) -> List[np.ndarray]:
        """
        Creates four depth images in every direction around the point
        for which the costmap should be created. The depth images are converted
        to metre, meaning that every entry in the depth images represents the
        distance to the next object in metre.

        :return: A list of four depth images, the images are represented as 2D arrays.
        """
        images = []
        camera_pose = self.origin

        self.move_target_and_robot_far_away()

        with UseProspectionWorld():
            origin_copy = self.origin.copy()
            origin_copy.position.y += 1
            images.append(
                self.world.get_images_for_target(origin_copy, camera_pose, size=self.size)[1])

            origin_copy = self.origin.copy()
            origin_copy.position.x -= 1
            images.append(self.world.get_images_for_target(origin_copy, camera_pose, size=self.size)[1])

            origin_copy = self.origin.copy()
            origin_copy.position.y -= 1
            images.append(self.world.get_images_for_target(origin_copy, camera_pose, size=self.size)[1])

            origin_copy = self.origin.copy()
            origin_copy.position.x += 1
            images.append(self.world.get_images_for_target(origin_copy, camera_pose, size=self.size)[1])

        self.return_target_and_robot_to_their_original_position()

        if not World.current_world.conf.depth_images_are_in_meter:
            for i in range(0, 4):
                images[i] = self._depth_buffer_to_meter(images[i])
        return images

    def _depth_buffer_to_meter(self, buffer: np.ndarray) -> np.ndarray:
        """
        Converts the depth images generated by the World to represent
        each position in metre.

        :return: The depth image in metre
        """
        near = 0.2
        far = 100
        return far * near / (far - (far - near) * buffer)

    def _generate_map(self):
        """
        This method generates the resulting density map by using the algorithm explained
        in Lorenz Mösenlechners `PhD Thesis (page 178) <https://mediatum.ub.tum.de/doc/1239461/1239461.pdf>`_
        The resulting map is then saved to :py:attr:`self.map`
        """
        depth_imgs = self._create_images()
        # A 2D array where every cell contains the arctan2 value with respect to
        # the middle of the array. Additionally, the interval is shifted such that
        # it is between 0 and 2pi
        tan = np.arctan2(np.mgrid[-int(self.size / 2): int(self.size / 2), -int(self.size / 2): int(self.size / 2)][0],
                         np.mgrid[-int(self.size / 2): int(self.size / 2), -int(self.size / 2): int(self.size / 2)][
                             1]) + np.pi
        res = np.zeros(tan.shape)

        # Just for completion, since the res array has zeros in every position this
        # operation is not necessary.
        # res[np.logical_and(tan <= np.pi * 0.25, tan >= np.pi * 1.75)] = 0

        # Creates a 2D array which contains the index of the depth image for every
        # coordinate
        res[np.logical_and(tan >= np.pi * 1.25, tan <= np.pi * 1.75)] = 3
        res[np.logical_and(tan >= np.pi * 0.75, tan < np.pi * 1.25)] = 2
        res[np.logical_and(tan >= np.pi * 0.25, tan < np.pi * 0.75)] = 1

        indices = np.dstack(np.mgrid[0: self.size, 0: self.size])
        depth_indices = np.zeros(indices.shape)
        # x-value of index: res == n, :1
        # y-value of index: res == n, 1:2

        # (y, size-x-1) for index between 1.25 pi and 1.75 pi
        depth_indices[res == 3, :1] = indices[res == 3, 1:2]
        depth_indices[res == 3, 1:2] = self.size - indices[res == 3, :1] - 1

        # (size-x-1, y) for index between 0.75 pi and 1.25 pi
        depth_indices[res == 2, :1] = self.size - indices[res == 2, :1] - 1
        depth_indices[res == 2, 1:2] = indices[res == 2, 1:2]

        # (size-y-1, x) for index between 0.25 pi and 0.75 pi
        depth_indices[res == 1, :1] = self.size - indices[res == 1, 1:2] - 1
        depth_indices[res == 1, 1:2] = indices[res == 1, :1]

        # (x, y) for index between 0.25 pi and 1.75 pi
        depth_indices[res == 0, :1] = indices[res == 0, :1]
        depth_indices[res == 0, 1:2] = indices[res == 0, 1:2]

        # Convert back to origin in the middle of the costmap
        depth_indices[:, :, :1] -= self.size / 2
        depth_indices[:, :, 1:2] = np.absolute(self.size / 2 - depth_indices[:, :, 1:2])

        # Sets the y index for the coordinates of the middle of the costmap to 1,
        # the computed value is 0 which would cause an error in the next step where
        # the calculation divides the x coordinates by the y coordinates
        depth_indices[int(self.size / 2), int(self.size / 2), 1] = 1

        # Calculate columns for the respective position in the costmap
        columns = np.around(((depth_indices[:, :, :1] / depth_indices[:, :, 1:2])
                             * (self.size / 2)) + self.size / 2).reshape((self.size, self.size)).astype('int16')

        # An array with size * size that contains the euclidean distance to the
        # origin (in the middle of the costmap) from every cell
        distances = np.maximum(np.linalg.norm(np.dstack(np.mgrid[-int(self.size / 2): int(self.size / 2), \
                                                        -int(self.size / 2): int(self.size / 2)]), axis=2), 0.001)

        # Row ranges
        # Calculation of the ranges of coordinates in the row which have to be
        # taken into account. The range is from r_min to r_max.
        # These are two arrays with shape: size*size, the r_min constrains the beginning
        # of the range for every coordinate and r_max contains the end for each
        # coordinate
        r_min = (np.arctan((self.min_height - self.origin.position.z) / distances) * self.size) + self.size / 2
        r_max = (np.arctan((self.max_height - self.origin.position.z) / distances) * self.size) + self.size / 2

        r_min = np.minimum(np.around(r_min), self.size - 1).astype('int16')
        r_max = np.minimum(np.around(r_max), self.size - 1).astype('int16')

        rs = np.dstack((r_min, r_max + 1)).reshape((self.size ** 2, 2))
        r = np.arange(self.size)
        # Calculates a mask from the r_min and r_max values. This mask is for every
        # coordinate respectively and determines which values from the computed column
        # of the depth image should be taken into account for the costmap.
        # A Mask of a single coordinate has the length of the column of the depth image
        # and together with the computed column at this coordinate determines which
        # values of the depth image make up the value of the visibility costmap at this
        # point.
        mask = ((rs[:, 0, None] <= r) & (rs[:, 1, None] > r)).reshape((self.size, self.size, self.size))

        values = np.zeros((self.size, self.size))
        map = np.zeros((self.size, self.size))
        # This is done to iterate over the depth images one at a time
        for i in range(4):
            row_masks = mask[res == i].T
            # This statement does several things, first it takes the values from
            # the depth image for this quarter of the costmap. The values taken are
            # the complete columns of the depth image (which where computed beforehand)
            # and checks if the values in them are greater than the distance to the
            # respective coordinates. This does not take the row ranges into account.
            values = depth_imgs[i][:, columns[res == i].flatten()] > \
                     np.tile(distances[res == i][:, None], (1, self.size)).T * self.resolution
            # This applies the created mask of the row ranges to the values of
            # the columns which are compared in the previous statement
            masked = np.ma.masked_array(values, mask=~row_masks)
            # The calculated values are added to the costmap
            map[res == i] = np.sum(masked, axis=0)
        map /= np.max(map)
        # Weird flipping shit so that the map fits the orientation of the visualization.
        # the costmap in itself is consistent and just needs to be flipped to fit the world coordinate system
        map = np.flip(map, axis=0)
        map = np.flip(map)
        self.map = map


class GaussianCostmap(Costmap):
    """
    Gaussian Costmaps are 2D gaussian distributions around the origin with the given mean and sigma
    """

    def __init__(self, mean: int, sigma: float, resolution: Optional[float] = 0.02,
                 origin: Optional[PoseStamped] = None):
        """
        This Costmap creates a 2D gaussian distribution around the origin with
        the specified size.

        :param mean: The mean input for the gaussian distribution, this also specifies
            the length of the side of the resulting costmap. The costmap is Created
            as a square.
        :param sigma: The sigma input for the gaussian distribution.
        :param resolution: The resolution of the costmap, this specifies how much
            meter a pixel represents.
        :param origin: The origin of the costmap around which it will be created.
        """
        self.gau: np.ndarray = self._gaussian_window(mean, sigma)
        self.map: np.ndarray = np.outer(self.gau, self.gau)
        self.size: float = mean
        self.origin: PoseStamped = PoseStamped.from_list() if not origin else origin
        Costmap.__init__(self, resolution, mean, mean, self.origin, self.map)

    def _gaussian_window(self, mean: int, std: float) -> np.ndarray:
        """
        This method creates a window of values with a gaussian distribution of
        size "mean" and standart deviation "std".
        Code from `Scipy <https://github.com/scipy/scipy/blob/v0.14.0/scipy/signal/windows.py#L976>`_
        """
        n = np.arange(0, mean) - (mean - 1.0) / 2.0
        sig2 = 2 * std * std
        w = np.exp(-n ** 2 / sig2)
        return w


class SemanticCostmap(Costmap):
    """
    Semantic Costmaps represent a 2D distribution over a link of an Object. An example of this would be a Costmap for a
    table surface.
    """

    def __init__(self, obj: Object, link_name: str, resolution: float = 0.02, world: Optional[World] = None):
        """
        Creates a semantic costmap for the given parameter. The semantic costmap will be on top of the link of the given
        Object.

        :param obj: The object of which the link is a part
        :param link_name: The link name, as stated in the description of the object
        :param resolution: Resolution of the final costmap (how much meters one pixel represents)
        :param world: The World from which the costmap should be created
        """
        self.world: World = world if world else World.current_world
        self.object: Object = obj
        self.link: Link = obj.get_link(link_name)
        self.resolution: float = resolution
        self.origin: PoseStamped = obj.get_link_pose(link_name)
        self.height: int = 0
        self.width: int = 0
        self.map: np.ndarray = []
        self.generate_map()

        Costmap.__init__(self, resolution, self.height, self.width, self.origin, self.map)

    def get_edges_map(self, margin_in_meters: float, horizontal_only: bool = False) -> Costmap:
        """
        Return a Costmap with only the edges of the original Costmap marked as possible positions.

        :param margin_in_meters: The edge thickness in meters that should be marked as possible positions.
        :param horizontal_only: If True only the horizontal edges will be marked as possible positions.
        :return: The modified Costmap.
        """
        mask = np.zeros(self.map.shape)
        edge_tolerance = int(margin_in_meters / self.resolution)
        mask[:edge_tolerance] = 1
        mask[-edge_tolerance:] = 1
        if not horizontal_only:
            mask[:, :edge_tolerance] = 1
            mask[:, -edge_tolerance:] = 1
        return Costmap(self.resolution, self.height, self.width, self.origin, mask)

    def generate_map(self) -> None:
        """
        Generates the semantic costmap according to the provided parameters. To do this the axis aligned bounding box (AABB)
        for the link name will be used. Height and width of the final Costmap will be the x and y sizes of the AABB.
        """
        min_p, max_p = self.get_aabb_for_link().get_min_max_points()
        self.height = int((max_p.x - min_p.x) // self.resolution)
        self.width = int((max_p.y - min_p.y) // self.resolution)
        self.map = np.ones((self.height, self.width))

    def get_aabb_for_link(self) -> AxisAlignedBoundingBox:
        """
        :return: The original untransformed (doesn't take the current pose of the link into consideration, since only
        the size is important here not the pose) axis aligned bounding box (AABB) of the link provided when creating
         this costmap.
        """
        return self.link.get_axis_aligned_bounding_box_from_geometry()


class AlgebraicSemanticCostmap(SemanticCostmap):
    """
    Class for a semantic costmap that is based on an algebraic set-description of the valid area.
    """
    x: Continuous = Continuous("x")
    """
    The variable for height.
    """

    y: Continuous = Continuous("y")
    """
    The variable for width.
    """

    original_valid_area: Optional[SimpleEvent]
    """
    The original rectangle of the valid area.
    """

    valid_area: Optional[Event]
    """
    A description of the valid positions as set.
    """

    number_of_samples: int
    """
    The number of samples to generate for the iter.
    """

    def __init__(self, object, urdf_link_name, world=None, number_of_samples=1000):
        super().__init__(object, urdf_link_name, world=world)
        self.number_of_samples = number_of_samples

    def check_valid_area_exists(self):
        assert self.valid_area is not None, ("The map has to be created before semantics can be applied. "
                                             "Call 'generate_map first'")

    def left(self, margin=0.) -> Event:
        """
        Create an event left of the origins Y-Coordinate.
        :param margin: The margin of the events left bound.
        :return: The left event.
        """
        self.check_valid_area_exists()
        y_origin = self.origin.position.y
        left = self.original_valid_area[self.y].simple_sets[0].lower
        left += margin
        event = SimpleEvent(
            {self.x: reals(), self.y: random_events.interval.open(left, y_origin)}).as_composite_set()
        return event

    def right(self, margin=0.) -> Event:
        """
        Create an event right of the origins Y-Coordinate.
        :param margin: The margin of the events right bound.
        :return: The right event.
        """
        self.check_valid_area_exists()
        y_origin = self.origin.position.y
        right = self.original_valid_area[self.y].simple_sets[0].upper
        right -= margin
        event = SimpleEvent({self.x: reals(), self.y: closed_open(y_origin, right)}).as_composite_set()
        return event

    def top(self, margin=0.) -> Event:
        """
        Create an event above the origins X-Coordinate.
        :param margin: The margin of the events upper bound.
        :return: The top event.
        """
        self.check_valid_area_exists()
        x_origin = self.origin.position.x
        top = self.original_valid_area[self.x].simple_sets[0].upper
        top -= margin
        event = SimpleEvent(
            {self.x: random_events.interval.closed_open(x_origin, top), self.y: reals()}).as_composite_set()
        return event

    def bottom(self, margin=0.) -> Event:
        """
        Create an event below the origins X-Coordinate.
        :param margin: The margin of the events lower bound.
        :return: The bottom event.
        """
        self.check_valid_area_exists()
        x_origin = self.origin.position.x
        lower = self.original_valid_area[self.x].simple_sets[0].lower
        lower += margin
        event = SimpleEvent(
            {self.x: random_events.interval.open(lower, x_origin), self.y: reals()}).as_composite_set()
        return event

    def inner(self, margin=0.2):
        min_x = self.original_valid_area[self.x].simple_sets[0].lower
        max_x = self.original_valid_area[self.x].simple_sets[0].upper
        min_y = self.original_valid_area[self.y].simple_sets[0].lower
        max_y = self.original_valid_area[self.y].simple_sets[0].upper

        min_x += margin
        max_x -= margin
        min_y += margin
        max_y -= margin

        inner_event = SimpleEvent({self.x: closed(min_x, max_x),
                                   self.y: closed(min_y, max_y)}).as_composite_set()
        return inner_event

    def border(self, margin=0.2):
        return ~self.inner(margin)

    def generate_map(self) -> None:
        super().generate_map()
        valid_area = None
        for rectangle in self.partitioning_rectangles():
            # rectangle.scale(1/self.resolution, 1/self.resolution)
            rectangle.translate(self.origin.position.x, self.origin.position.y)
            rectangle_event = SimpleEvent({self.x: closed(rectangle.x_lower, rectangle.x_upper),
                                           self.y: closed(rectangle.y_lower, rectangle.y_upper)}).as_composite_set()
            if valid_area is None:
                valid_area = rectangle_event
            else:
                valid_area |= rectangle_event

        assert len(valid_area.simple_sets) == 1, ("The map at the basis of a Semantic costmap must be an axis aligned"
                                                  "bounding box")
        self.valid_area = valid_area
        self.original_valid_area = self.valid_area.simple_sets[0]

    def as_distribution(self) -> ProbabilisticCircuit:
        model = uniform_measure_of_event(self.valid_area)
        return model

    def sample_to_pose(self, sample: np.ndarray) -> PoseStamped:
        """
        Convert a sample from the costmap to a pose.

        :param sample: The sample to convert
        :return: The pose corresponding to the sample
        """
        x = sample[0]
        y = sample[1]
        position = [x, y, self.origin.position.z]
        angle = np.arctan2(position[1] - self.origin.position.y, position[0] - self.origin.position.x) + np.pi
        orientation = list(quaternion_from_euler(0, 0, angle, axes="sxyz"))
        return PoseStamped.from_list(position, orientation, self.origin.frame_id)

    def __iter__(self) -> Iterator[PoseStamped]:
        model = self.as_distribution()
        samples = model.sample(self.number_of_samples)
        for sample in samples:
            yield self.sample_to_pose(sample)


cmap = colors.ListedColormap(['white', 'black', 'green', 'red', 'blue'])


# Mainly used for debugging
# Data is 2d array
def plot_grid(data: np.ndarray) -> None:
    """
    An auxiliary method only used for debugging, it will plot a 2D numpy array using MatplotLib.
    """
    rows = data.shape[0]
    cols = data.shape[1]
    fig, ax = plt.subplots()
    ax.imshow(data, cmap=cmap)
    # draw gridlines
    # ax.grid(which='major', axis='both', linestyle='-', rgba_color='k', linewidth=1)
    ax.set_xticks(np.arange(0.5, rows, 1));
    ax.set_yticks(np.arange(0.5, cols, 1));
    plt.tick_params(axis='both', labelsize=0, length=0)
    # fig.set_size_inches((8.5, 11), forward=False)
    # plt.savefig(saveImageName + ".png", dpi=500)
    plt.show()



from __future__ import annotations

import sys
import logging

import numpy as np
from pytransform3d.transform_manager import TransformManager
from pytransform3d.transformations import transform_from_pq
from pytransform3d.rotations import quaternion_from_matrix

from .ros import Time

from .tf_transformations import quaternion_matrix

if 'world' in sys.modules:
    logging.warning("(publisher) Make sure that you are not loading this module from pycram.world.")

from .datastructures.pose import PoseStamped, TransformStamped, Pose, Header, Transform, Vector3, Quaternion
from typing_extensions import List, Optional, Union, Iterable, TYPE_CHECKING, Tuple, Dict, Hashable

if TYPE_CHECKING:
    from .world_concepts.world_object import Object
    from .datastructures.world import World


class LocalTransformer(TransformManager):
    _instance = None
    prospection_prefix: str = "prospection/"

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized: return
        super().__init__()
        self.world: World = None
        # TODO: Ask Jonas if this is still needed
        self.prospection_world: World = None
        self._initialized = True
        self.add_transform("map", "map", transform_from_pq((0, 0, 0, 1, 0, 0, 0)))

    def transform_to_object_frame(self, pose: PoseStamped,
                                  world_object: Object, link_name: str = None) -> Union[
        PoseStamped, None]:
        """
        Transforms the given PoseStamped to the object coordinate frame

        :param pose: The Pose that should be transformed
        :param world_object: The world object to which coordinate frame should be transformed
        :param link_name: An optional link name of the object
        :return: The transformed PoseStamped
        """
        if link_name:
            target_frame = world_object.get_link_tf_frame(link_name)
        else:
            target_frame = world_object.tf_frame
        return self.transform_pose(pose, target_frame)

    def update_transforms_for_objects(self, objects: List[Object]) -> None:
        """
        Updates the transforms for objects affected by the transformation. The objects are identified by their names.

        :param objects: List of objects for which the transforms should be updated
        """
        for obj in objects:
            obj.update_link_transforms()
            obj.update_transform()
            self.update_transforms([obj.pose.to_transform_stamped(obj.tf_frame)])

    def transform_pose(self, pose: PoseStamped, target_frame: str) -> Optional[PoseStamped]:
        """
        Transforms the given PoseStamped to an arbitrary coordinate frame

        :param pose: The pose that should be transformed
        :param target_frame: The name of the target frame
        :return: The transformed PoseStamped in target frame
        """
        self._check_pose_integrity(pose)

        objects = list(filter(None, map(self.get_object_from_frame, [pose.frame_id, target_frame])))
        self.update_transforms_for_objects(objects)

        source_frame = pose.header.frame_id

        wxyz = self.xyzw_to_wxyz(pose.pose.orientation.to_list())

        pose_matrix = transform_from_pq(np.hstack((np.array(pose.pose.position.to_list()),
                                                   np.array(wxyz))))

        transform_matrix = self.get_transform(target_frame, source_frame)

        new_pose = transform_matrix @ pose_matrix

        return_pose = PoseStamped(pose=Pose.from_matrix(new_pose), header=Header(frame_id=target_frame))
        return return_pose

    def get_object_from_frame(self, frame: str) -> Optional[Object]:
        """
        Get the name of the object that is associated with the given frame.

        :param frame: The frame for which the object name should be returned
        :return: The name of the object associated with the frame
        """
        if frame == "map":
            return None
        world = self.get_world_from_frame(frame)
        found_objects = [obj for obj in world.objects if frame == obj.tf_frame]
        return found_objects[0] if len(found_objects) > 0 else self.get_object_from_link_frame(frame)

    @staticmethod
    def translate_pose_along_local_axis(pose: PoseStamped, axis: List, distance: float) -> PoseStamped:
        """
        Translate a pose along a given 3d vector (axis) by a given distance. The axis is given in the local coordinate
        frame of the pose. The axis is normalized and then scaled by the distance.

        :param pose: The pose that should be translated
        :param axis: The local axis along which the translation should be performed
        :param distance: The distance by which the pose should be translated

        :return: The translated pose
        """
        normalized_translation_vector = np.array(axis) / np.linalg.norm(axis)

        rot_matrix = quaternion_matrix(pose.orientation.to_list())[:3, :3]
        translation_in_world = rot_matrix @ normalized_translation_vector
        scaled_translation_vector = np.array(pose.position.to_list()) + translation_in_world * distance

        return PoseStamped.from_list(list(scaled_translation_vector), pose.orientation.to_list(), pose.frame_id)

    def get_object_from_link_frame(self, link_frame: str) -> Optional[Object]:
        """
        Get the name of the object that is associated with the given link frame.

        :param link_frame: The frame of the link for which the object name should be returned
        :return: The name of the object associated with the link frame
        """
        world = self.get_world_from_frame(link_frame)
        found_objects = [obj for obj in world.objects for link in obj.links.values()
                         if link_frame in (link.name, link.tf_frame)]
        return found_objects[0] if len(found_objects) > 0 else None

    def get_world_from_frame(self, frame: str) -> World:
        """
        Get the world that is associated with the given frame name.

        :param frame: The frame name.
        """
        return self.prospection_world if self.prospection_prefix in frame else self.world

    def lookup_transform_from_source_to_target_frame(self, source_frame: str, target_frame: str,
                                                     time: Optional[Time] = None) -> TransformStamped:
        pass

    def update_transforms(self, transforms: Iterable[TransformStamped], time: Time = None) -> None:
        """
       Updates transforms by updating the time stamps of the header of each transform. If no time is given the current
       time is used.
       """
        for transform in transforms:
            wxyz = self.xyzw_to_wxyz(transform.rotation.to_list())
            pt_transform = transform_from_pq(np.hstack((np.array(transform.translation.to_list()),
                                                        np.array(wxyz))))
            self.add_transform(transform.header.frame_id, transform.child_frame_id, pt_transform)

    @staticmethod
    def xyzw_to_wxyz(xyzw: List[float]) -> List[float]:
        """
        Convert a quaternion from XYZW to WXYZ format.

        :param xyzw: The quaternion in XYZW format.
        """
        return [xyzw[3], *xyzw[:3]]

    def get_all_frames(self) -> Dict[Tuple[Hashable, Hashable], Transform]:
        return {key: Transform.from_matrix(value) for key, value in self.transforms.items()}

    @staticmethod
    def _check_pose_integrity(pose: PoseStamped) -> None:
        """
        Check the integrity of the pose. This is a placeholder for any integrity checks that might be needed.
        """
        if not isinstance(pose, PoseStamped):
            raise TypeError("The pose must be of type PoseStamped.")
        elif not isinstance(pose.position, Vector3):
            raise TypeError("The position must be of type Vector3.")
        elif not isinstance(pose.orientation, Quaternion):
            raise TypeError("The orientation must be of type Quaternion.")



import rclpy.logging
from typing_extensions import TYPE_CHECKING

from . import node
import os

from ...datastructures.enums import LoggerLevel

# Default logger format
os.environ['RCUTILS_CONSOLE_OUTPUT_FORMAT'] = "[{severity}] [{name}]: {message} ({function_name}() at {file_name}:{line_number})"
# os.environ['RCUTILS_COLORIZED_OUTPUT'] = "[{file_name}:{line_number}:{function_name}] {message}"

# Set the logger level to INFO
node.get_logger().set_level(rclpy.logging.LoggingSeverity.INFO)

def set_logger_level(level: LoggerLevel):
    """
    Set the logger level for the pycram logger. Possible values are: DEBUG, INFO, WARN, ERROR, FATAL.

    :param level: The level to set the logger to.
    """
    lev = rclpy.logging.LoggingSeverity.INFO
    if level == LoggerLevel.DEBUG:
        lev = rclpy.logging.LoggingSeverity.DEBUG
    elif level == LoggerLevel.INFO:
        lev = rclpy.logging.LoggingSeverity.INFO
    elif level == LoggerLevel.WARN:
        lev = rclpy.logging.LoggingSeverity.WARN
    elif level == LoggerLevel.ERROR:
        lev = rclpy.logging.LoggingSeverity.ERROR
    elif level == LoggerLevel.FATAL:
        lev = rclpy.logging.LoggingSeverity.FATAL

    node.get_logger().set_level(lev)

def set_logger_format(logger_format: str):
    """
    Set the logger format for the pycram logger. The format can contain the following placeholders:
    - {severity}
    - {name}
    - {message}
    - {function_name}
    - {file_name}
    - {line_number}
    The format is set to the environment variable RCUTILS_CONSOLE_OUTPUT_FORMAT. Therefore, the format is not only
    applied to the pycram logger but to all loggers in the environment.

    :param logger_format: The format to set the logger to.
    """
    os.environ['RCUTILS_CONSOLE_OUTPUT_FORMAT'] = logger_format

def logwarn(message: str):
    node.get_logger().warn(message)

def loginfo(message: str):
    node.get_logger().info(message)

def logerr(message: str):
    node.get_logger().error(message)

def logdebug(message: str):
    node.get_logger().debug(message)

def logwarn_once(message: str):
    node.get_logger().warn(message, once=True)

def loginfo_once(message: str):
    node.get_logger().info(message, once=True)

def logerr_once(message: str):
    node.get_logger().error(message, once=True)

def logdebug_once(message: str):
    node.get_logger().debug(message, once=True)


import logging
import os
import shutil
from pathlib import Path
from time import sleep

import numpy as np
from ..tf_transformations import quaternion_matrix
from typing_extensions import List, Dict, Optional, Union, Tuple, Callable, Type

import pycrap
from pycrap import PhysicalObject
from .multiverse_communication.client_manager import MultiverseClientManager
from .multiverse_communication.clients import MultiverseController, MultiverseReader, MultiverseWriter, MultiverseAPI
from ..config.multiverse_conf import MultiverseConfig
from ..datastructures.dataclasses import Color, ContactPointsList, ContactPoint, RayResult
from ..datastructures.enums import WorldMode, JointType, MultiverseBodyProperty, MultiverseJointPosition, \
    MultiverseJointCMD
from ..datastructures.pose import PoseStamped
from ..datastructures.world import World
from ..datastructures.world_entity import PhysicalBody
from ..description import Link, Joint
from ..object_descriptors.generic import ObjectDescription as GenericObjectDescription
from ..object_descriptors.mjcf import ObjectDescription as MJCF, PrimitiveObjectFactory
from ..robot_description import RobotDescription
from ..ros import  logwarn
from ..utils import RayTestUtils, wxyz_to_xyzw, xyzw_to_wxyz, adjust_camera_pose_based_on_target
from ..validation.goal_validator import validate_object_pose, validate_multiple_joint_positions, \
    validate_joint_position, validate_multiple_object_poses
from ..world_concepts.constraints import Constraint
from ..world_concepts.world_object import Object


class Multiverse(World):
    """
    This class implements an interface between Multiverse and PyCRAM.
    """

    conf: MultiverseConfig = MultiverseConfig
    """
    The Multiverse configuration.
    """

    supported_joint_types = (JointType.REVOLUTE, JointType.CONTINUOUS, JointType.PRISMATIC)
    """
    A Tuple for the supported pycram joint types in Multiverse.
    """

    added_multiverse_resources: bool = False
    """
    A flag to check if the multiverse resources have been added.
    """

    Object.extension_to_description_type[MJCF.get_file_extension()] = MJCF
    """
    Add the MJCF description extension to the extension to description type mapping for the objects.
    """

    def __init__(self, is_prospection: Optional[bool] = False,
                 clear_cache: bool = False):
        """
        Initialize the Multiverse Socket and the PyCram World.

        :param is_prospection: Whether the world is prospection or not.
        :param clear_cache: Whether to clear the cache or not.
        """

        self.latest_save_id: Optional[int] = None
        self.saved_simulator_states: Dict = {}
        self.make_sure_multiverse_resources_are_added(clear_cache=clear_cache)

        self.simulation = self.conf.prospection_world_prefix if is_prospection else "belief_state"
        self.client_manager = MultiverseClientManager(self.conf.simulation_wait_time_factor)
        self._init_clients(is_prospection=is_prospection)

        World.__init__(self, mode=WorldMode.DIRECT, is_prospection=is_prospection)

        self._init_constraint_and_object_id_name_map_collections()

        self.ray_test_utils = RayTestUtils(self.ray_test_batch, self.object_id_to_name)

        self.is_paused: bool = False

        if not self.is_prospection_world:
            self._spawn_floor()

        if self.conf.use_static_mode:
            self.pause_simulation()

    def _init_clients(self, is_prospection: bool = False):
        """
        Initialize the Multiverse clients that will be used to communicate with the Multiverse server.
        Each client is responsible for a specific task, e.g. reading data from the server, writing data to the serve,
         calling the API, or controlling the robot joints.

        :param is_prospection: Whether the world is prospection or not.
        """
        self.reader: MultiverseReader = self.client_manager.create_reader(
            is_prospection_world=is_prospection)
        self.writer: MultiverseWriter = self.client_manager.create_writer(
            self.simulation,
            is_prospection_world=is_prospection)
        self.api_requester: MultiverseAPI = self.client_manager.create_api_requester(
            self.simulation,
            is_prospection_world=is_prospection)
        if self.conf.use_controller:
            self.joint_controller: MultiverseController = self.client_manager.create_controller(
                is_prospection_world=is_prospection)

    def _init_constraint_and_object_id_name_map_collections(self):
        self.last_object_id: int = -1
        self.last_constraint_id: int = -1
        self.constraints: Dict[int, Constraint] = {}
        self.object_name_to_id: Dict[str, int] = {}
        self.object_id_to_name: Dict[int, str] = {}

    def _init_world(self, mode: WorldMode):
        pass

    @classmethod
    def make_sure_multiverse_resources_are_added(cls, clear_cache: bool = False) -> None:
        """
        Add the multiverse resources to the pycram world resources, and change the data directory and cache manager.

        :param clear_cache: Whether to clear the cache or not.
        """
        if not cls.added_multiverse_resources:
            if clear_cache:
                World.cache_manager.clear_cache()
            World.add_resource_path(cls.conf.resources_path, prepend=True)
            World.change_cache_dir_path(cls.conf.resources_path)
            cls.added_multiverse_resources = True

    def remove_multiverse_resources(self):
        """
        Remove the multiverse resources from the pycram world resources.
        """
        if self.added_multiverse_resources:
            World.remove_resource_path(self.conf.resources_path)
            World.change_cache_dir_path(self.conf.cache_dir)
            self.added_multiverse_resources = False

    def _spawn_floor(self):
        """
        Spawn the plane in the simulator.
        """
        self.floor = Object("floor", pycrap.Floor, "plane.urdf",
                            world=self)

    def pause_simulation(self) -> None:
        """
        Pause the simulation.
        """
        self.api_requester.pause_simulation()
        self.is_paused = True

    def unpause_simulation(self) -> None:
        """
        Unpause the simulation.
        """
        self.api_requester.unpause_simulation()
        self.is_paused = False

    def load_generic_object_and_get_id(self, description: GenericObjectDescription,
                                       pose: Optional[PoseStamped] = None) -> int:
        save_path = os.path.join(self.cache_manager.cache_dir, description.name + ".xml")
        object_factory = PrimitiveObjectFactory(description.name, description.links[0].geometry, save_path)
        object_factory.build_shape()
        object_factory.export_to_mjcf(save_path)
        return self.load_object_and_get_id(description.name, pose, pycrap.PhysicalObject)

    def get_images_for_target(self, target_pose: PoseStamped,
                              cam_pose: PoseStamped,
                              size: int = 256,
                              camera_min_distance: float = 0.1,
                              camera_max_distance: int = 3,
                              plot: bool = False) -> List[np.ndarray]:
        """
        Uses ray test to get the images for the target object. (target_pose is currently not used)
        """
        camera_description = RobotDescription.current_robot_description.get_default_camera()
        camera_frame = RobotDescription.current_robot_description.get_camera_frame(World.robot.name)
        adjusted_cam_pose = adjust_camera_pose_based_on_target(cam_pose, target_pose, camera_description)
        return self.ray_test_utils.get_images_for_target(adjusted_cam_pose, camera_description, camera_frame,
                                                         size, camera_min_distance, camera_max_distance, plot)

    @staticmethod
    def get_joint_position_name(joint: Joint) -> MultiverseJointPosition:
        """
        Get the attribute name of the joint position in the Multiverse from the pycram joint type.

        :param joint: The joint.
        """
        return MultiverseJointPosition.from_pycram_joint_type(joint.type)

    def spawn_robot_with_controller(self, name: str, pose: PoseStamped) -> None:
        """
        Spawn the robot in the simulator.

        :param name: The name of the robot.
        :param pose: The pose of the robot.
        """
        actuator_joint_commands = {
            actuator_name: [self.get_joint_cmd_name(self.robot_description.joint_types[joint_name]).value]
            for joint_name, actuator_name in self.robot_joint_actuators.items()
        }
        self.joint_controller.init_controller(actuator_joint_commands)
        self.writer.spawn_robot_with_actuators(name, actuator_joint_commands)
        if not pose.almost_equal(PoseStamped()):
            goal = self.robot.get_mobile_base_joint_goal(pose)
            self.set_multiple_joint_positions(goal)

    def load_object_and_get_id(self, name: Optional[str] = None,
                               pose: Optional[PoseStamped] = None,
                               obj_type: Optional[Type[PhysicalObject]] = None) -> int:
        """
        Spawn the object in the simulator and return the object id. Object name has to be unique and has to be same as
        the name of the object in the description file.

        :param name: The name of the object to be loaded.
        :param pose: The pose of the object.
        :param obj_type: The type of the object.
        """
        if pose is None:
            pose = PoseStamped()

        # Do not spawn objects with type environment as they should be already present in the simulator through the
        # multiverse description file (.muv file).
        obj_type = pycrap.PhysicalObject if obj_type is None else obj_type
        if not (issubclass(obj_type, pycrap.Location) or issubclass(obj_type, pycrap.Floor)):
            self.spawn_object(name, obj_type, pose)

        return self._update_object_id_name_maps_and_get_latest_id(name)

    def spawn_object(self, name: str, object_type: Type[PhysicalObject], pose: PoseStamped) -> None:
        """
        Spawn the object in the simulator.

        :param name: The name of the object.
        :param object_type: The type of the object.
        :param pose: The pose of the object.
        """
        if issubclass(object_type, pycrap.Robot):
            if self.conf.use_controller:
                self.spawn_robot_with_controller(name, pose)
            else:
                self.spawn_robot(name, pose)
        else:
            self._set_body_pose(name, pose)

    def spawn_robot(self, name: str, pose: PoseStamped) -> None:
        """
        Spawn the robot in the simulator.

        :param name: The name of the robot.
        :param pose: The pose of the robot.
        """
        self._set_body_pose(name, PoseStamped())
        self.robot.set_mobile_robot_pose(pose)

    def _update_object_id_name_maps_and_get_latest_id(self, name: str) -> int:
        """
        Update the object id name maps and return the latest object id.

        :param name: The name of the object.
        :return: The latest object id.
        """
        self.last_object_id += 1
        self.object_name_to_id[name] = self.last_object_id
        self.object_id_to_name[self.last_object_id] = name
        return self.last_object_id

    def get_object_joint_names(self, obj: Object) -> List[str]:
        return [joint.name for joint in obj.description.joints if joint.type in self.supported_joint_types]

    def get_object_link_names(self, obj: Object) -> List[str]:
        return [link.name for link in obj.description.links]

    def get_link_position(self, link: Link) -> List[float]:
        return self.reader.get_body_position(link.name)

    def get_link_orientation(self, link: Link) -> List[float]:
        return self.reader.get_body_orientation(link.name)

    def get_multiple_link_positions(self, links: List[Link]) -> Dict[str, List[float]]:
        return self.reader.get_multiple_body_positions([link.name for link in links])

    def get_multiple_link_orientations(self, links: List[Link]) -> Dict[str, List[float]]:
        return self.reader.get_multiple_body_orientations([link.name for link in links])

    @validate_joint_position
    def _reset_joint_position(self, joint: Joint, joint_position: float) -> bool:
        if not self.is_paused and self.conf.use_controller and self.joint_has_actuator(joint):
            self._reset_joint_position_using_controller(joint, joint_position)
        else:
            self._set_multiple_joint_positions_without_controller({joint: joint_position})
        return True

    def _reset_joint_position_using_controller(self, joint: Joint, joint_position: float) -> bool:
        """
        Reset the position of a joint in the simulator using the controller.

        :param joint: The joint.
        :param joint_position: The position of the joint.
        :return: True if the joint position is reset successfully.
        """
        self.joint_controller.set_body_property(self.get_actuator_for_joint(joint),
                                                self.get_joint_cmd_name(joint.type),
                                                [joint_position])
        return True

    @validate_multiple_joint_positions
    def _set_multiple_joint_positions(self, joint_positions: Dict[Joint, float]) -> bool:
        """
        Set the positions of multiple joints in the simulator. Also check if the joint is controlled by an actuator
        and use the controller to set the joint position if the joint is controlled.

        :param joint_positions: The dictionary of joints and positions.
        :return: True if the joint positions are set successfully (this means that the joint positions are set without
         errors, but not necessarily that the joint positions are set to the specified values).
        """

        if not self.is_paused and self.conf.use_controller:
            controlled_joints = self.get_controlled_joints(list(joint_positions.keys()))
            if len(controlled_joints) > 0:
                controlled_joint_positions = {joint: joint_positions[joint] for joint in controlled_joints}
                self._set_multiple_joint_positions_using_controller(controlled_joint_positions)
                joint_positions = {joint: joint_positions[joint] for joint in joint_positions.keys()
                                   if joint not in controlled_joints}
        if len(joint_positions) > 0:
            self._set_multiple_joint_positions_without_controller(joint_positions)

        return True

    def get_controlled_joints(self, joints: Optional[List[Joint]] = None) -> List[Joint]:
        """
        Get the joints that are controlled by an actuator from the list of joints.

        :param joints: The list of joints to check.
        :return: The list of controlled joints.
        """
        joints = self.robot.joints if joints is None else joints
        return [joint for joint in joints if self.joint_has_actuator(joint)]

    def _set_multiple_joint_positions_without_controller(self, joint_positions: Dict[Joint, float]) -> None:
        """
        Set the positions of multiple joints in the simulator without using the controller.

        :param joint_positions: The dictionary of joints and positions.
        """
        joints_data = {joint.name: {self.get_joint_position_name(joint): [position]}
                       for joint, position in joint_positions.items()}
        self.writer.send_multiple_body_data_to_server(joints_data)

    def _set_multiple_joint_positions_using_controller(self, joint_positions: Dict[Joint, float]) -> bool:
        """
        Set the positions of multiple joints in the simulator using the controller.

        :param joint_positions: The dictionary of joints and positions.
        """
        controlled_joints_data = {self.get_actuator_for_joint(joint):
                                  {self.get_joint_cmd_name(joint.type): [position]}
                                  for joint, position in joint_positions.items()}
        self.joint_controller.send_multiple_body_data_to_server(controlled_joints_data)
        return True

    def _get_joint_position(self, joint: Joint) -> Optional[float]:
        joint_position_name = self.get_joint_position_name(joint)
        data = self.reader.get_body_data(joint.name, [joint_position_name])
        if data is not None:
            return data[joint_position_name.value][0]

    def _get_multiple_joint_positions(self, joints: List[Joint]) -> Optional[Dict[str, float]]:
        joint_names = [joint.name for joint in joints]
        data = self.reader.get_multiple_body_data(joint_names, {joint.name: [self.get_joint_position_name(joint)]
                                                                for joint in joints})
        if data is not None:
            return {name: list(value.values())[0][0] for name, value in data.items()}

    @staticmethod
    def get_joint_cmd_name(joint_type: JointType) -> MultiverseJointCMD:
        """
        Get the attribute name of the joint command in the Multiverse from the pycram joint type.

        :param joint_type: The pycram joint type.
        """
        return MultiverseJointCMD.from_pycram_joint_type(joint_type)

    def get_link_pose(self, link: Link) -> Optional[PoseStamped]:
        return self._get_body_pose(link.name)

    def get_multiple_link_poses(self, links: List[Link]) -> Dict[str, PoseStamped]:
        return self._get_multiple_body_poses([link.name for link in links])

    def get_object_pose(self, obj: Object) -> PoseStamped:
        if obj.is_an_environment:
            return PoseStamped()
        return self._get_body_pose(obj.name)

    def get_multiple_object_poses(self, objects: List[Object]) -> Dict[str, PoseStamped]:
        """
        Set the poses of multiple objects in the simulator. If the object is of type environment, the pose will be
        the default pose.

        :param objects: The list of objects.
        :return: The dictionary of object names and poses.
        """
        non_env_objects = [obj for obj in objects if not obj.is_an_environment]
        all_poses = self._get_multiple_body_poses([obj.name for obj in non_env_objects])
        all_poses.update({obj.name: PoseStamped() for obj in objects if obj.is_an_environment})
        return all_poses

    @validate_object_pose
    def reset_object_base_pose(self, obj: Object, pose: PoseStamped) -> bool:
        if obj.is_an_environment:
            return False

        if (obj.is_a_robot and
                RobotDescription.current_robot_description.virtual_mobile_base_joints is not None):
            obj.set_mobile_robot_pose(pose)
        else:
            self._set_body_pose(obj.name, pose)

        return True

    @validate_multiple_object_poses
    def reset_multiple_objects_base_poses(self, objects: Dict[Object, PoseStamped]) -> None:
        """
        Reset the poses of multiple objects in the simulator.

        :param objects: The dictionary of objects and poses.
        """
        for obj in objects.keys():
            if (obj.is_a_robot and
                    RobotDescription.current_robot_description.virtual_mobile_base_joints is not None):
                obj.set_mobile_robot_pose(objects[obj])
        objects = {obj: pose for obj, pose in objects.items()
                   if not obj.is_a_robot and not obj.is_an_environment}
        if len(objects) > 0:
            self._set_multiple_body_poses({obj.name: pose for obj, pose in objects.items()})

    def _set_body_pose(self, body_name: str, pose: PoseStamped) -> None:
        """
        Reset the pose of a body (object, link, or joint) in the simulator.

        :param body_name: The name of the body.
        :param pose: The pose of the body.
        """
        self._set_multiple_body_poses({body_name: pose})

    def _set_multiple_body_poses(self, body_poses: Dict[str, PoseStamped]) -> None:
        """
        Reset the poses of multiple bodies in the simulator.

        :param body_poses: The dictionary of body names and poses.
        """
        self.writer.set_multiple_body_poses({name: {MultiverseBodyProperty.POSITION: pose.position.to_list(),
                                                    MultiverseBodyProperty.ORIENTATION:
                                                        xyzw_to_wxyz(pose.orientation.to_list()),
                                                    MultiverseBodyProperty.RELATIVE_VELOCITY: [0.0] * 6}
                                             for name, pose in body_poses.items()})

    def _get_body_pose(self, body_name: str, wait: Optional[bool] = True) -> Optional[PoseStamped]:
        """
        Get the pose of a body in the simulator.

        :param body_name: The name of the body.
        :param wait: Whether to wait until the pose is received.
        :return: The pose of the body.
        """
        data = self.reader.get_body_pose(body_name, wait)
        return PoseStamped(data[MultiverseBodyProperty.POSITION.value],
                           wxyz_to_xyzw(data[MultiverseBodyProperty.ORIENTATION.value]))

    def _get_multiple_body_poses(self, body_names: List[str]) -> Dict[str, PoseStamped]:
        """
        Get the poses of multiple bodies in the simulator.

        :param body_names: The list of body names.
        """
        return self.reader.get_multiple_body_poses(body_names)

    def get_multiple_object_positions(self, objects: List[Object]) -> Dict[str, List[float]]:
        return self.reader.get_multiple_body_positions([obj.name for obj in objects])

    def get_object_position(self, obj: Object) -> List[float]:
        return self.reader.get_body_position(obj.name)

    def get_multiple_object_orientations(self, objects: List[Object]) -> Dict[str, List[float]]:
        return self.reader.get_multiple_body_orientations([obj.name for obj in objects])

    def get_object_orientation(self, obj: Object) -> List[float]:
        return self.reader.get_body_orientation(obj.name)

    def multiverse_reset_world(self):
        """
        Reset the world using the Multiverse API.
        """
        self.writer.reset_world()

    def disconnect_from_physics_server(self) -> None:
        MultiverseClientManager.stop_all_clients()

    def join_threads(self) -> None:
        self.reader.stop_thread = True
        self.reader.join()

    def _remove_visual_object(self, obj_id: int) -> bool:
        logwarn("Currently multiverse does not create visual objects")
        return False

    def remove_object_from_simulator(self, obj: Object) -> bool:
        if not obj.is_an_environment:
            self.writer.remove_body(obj.name)
            return True
        logwarn("Cannot remove environment objects")
        return False

    def add_constraint(self, constraint: Constraint) -> int:

        if constraint.type != JointType.FIXED:
            logging.error("Only fixed constraints are supported in Multiverse")
            raise ValueError

        if not self.conf.let_pycram_move_attached_objects:
            self.api_requester.attach(constraint)

        return self._update_constraint_collection_and_get_latest_id(constraint)

    def _update_constraint_collection_and_get_latest_id(self, constraint: Constraint) -> int:
        """
        Update the constraint collection and return the latest constraint id.

        :param constraint: The constraint to be added.
        :return: The latest constraint id.
        """
        self.last_constraint_id += 1
        self.constraints[self.last_constraint_id] = constraint
        return self.last_constraint_id

    def remove_constraint(self, constraint_id) -> None:
        constraint = self.constraints.pop(constraint_id)
        self.api_requester.detach(constraint)

    def perform_collision_detection(self) -> None:
        ...

    def get_body_contact_points(self, body: PhysicalBody) -> ContactPointsList:
        if isinstance(body, Object):
            return self.get_object_contact_points(body)
        else:
            multiverse_contact_points = self.api_requester.get_contact_points(body.name)
            contact_points = ContactPointsList([])
            for mcp in multiverse_contact_points:
                body_object, body_link = self.get_object_with_body_name(mcp.body_2)
                obj, obj_link = self.get_object_with_body_name(mcp.body_1)
                contact_points.append(ContactPoint(obj_link, body_link))
                contact_points[-1].normal_on_body_b = mcp.normal
                contact_points[-1].position_on_b = mcp.position
            return contact_points

    def get_object_contact_points(self, obj: Object, ignore_attached_objects: bool = True) -> ContactPointsList:
        """
        Note: Currently Multiverse only gets one contact point per contact objects.

        :param obj: The object.
        :param ignore_attached_objects: Whether to ignore the attached objects or not.
        :return: The contact points of the object.
        """
        multiverse_contact_points = self.api_requester.get_contact_points(obj.name)
        if ignore_attached_objects and len(obj.attachments) > 0:
            attached_objects_link_names = []
            for att_obj in obj.attachments.keys():
                attached_objects_link_names.extend(att_obj.link_names)
            multiverse_contact_points = [mcp for mcp in multiverse_contact_points if
                                         mcp.body_1 not in attached_objects_link_names]
        contact_points = ContactPointsList([])
        for mcp in multiverse_contact_points:
            body_object, body_link = self.get_object_with_body_name(mcp.body_2)
            obj_link = obj.links[mcp.body_1] if mcp.body_1 in obj.links.keys() else obj.root_link
            contact_points.append(ContactPoint(obj_link, body_link))
            contact_points[-1].normal_on_body_b = mcp.normal
            contact_points[-1].position_on_b = mcp.position
        return contact_points

    def get_object_with_body_name(self, body_name: str) -> Tuple[Optional[Object], Optional[Link]]:
        """
        Get the object with the body name in the simulator, the body name can be the name of the object or the link.

        :param body_name: The name of the body.
        :return: The object if found otherwise None, and the link if the body name is a link.
        """
        if body_name == "world":
            body_name = "floor"
        body_object = self.get_object_by_name(body_name)
        body_link = None
        if body_object is None:
            for obj in self.objects:
                for link in obj.links.values():
                    if link.name == body_name:
                        body_link = link
                        body_object = obj
                        break
        else:
            body_link = body_object.root_link
        return body_object, body_link

    @staticmethod
    def _get_normal_force_on_object_from_contact_force(obj: Object, contact_force: List[float]) -> float:
        """
        Get the normal force on an object from the contact force exerted by another object that is expressed in the
        world frame. Thus transforming the contact force to the object frame is necessary.

        :param obj: The object.
        :param contact_force: The contact force.
        :return: The normal force on the object.
        """
        obj_quat = obj.get_orientation_as_list()
        obj_rot_matrix = quaternion_matrix(obj_quat)[:3, :3]
        # invert the rotation matrix to get the transformation from world to object frame
        obj_rot_matrix = np.linalg.inv(obj_rot_matrix)
        contact_force_array = obj_rot_matrix @ np.array(contact_force).reshape(3, 1)
        return contact_force_array.flatten().tolist()[2]

    def get_contact_points_between_two_bodies(self, obj1: Object, obj2: Object) -> ContactPointsList:
        obj1_contact_points = self.get_object_contact_points(obj1)
        return obj1_contact_points.get_points_of_object(obj2)

    def _ray_test(self, from_position: List[float], to_position: List[float]) -> RayResult:
        return self.ray_test_batch([from_position], [to_position])[0]

    def _ray_test_batch(self, from_positions: List[List[float]],
                        to_positions: List[List[float]],
                        num_threads: int = 1,
                        return_distance: bool = False) -> List[RayResult]:
        """
        Note: Currently, num_threads is not used in Multiverse.
        """
        ray_results = self.api_requester.get_objects_intersected_with_rays(from_positions, to_positions)
        results: List[RayResult] = []
        for ray_result in ray_results:
            results.append(RayResult(-1))
            if ray_result.intersected:
                body_name = ray_result.body_name
                if body_name == "world":
                    results[-1].obj_id = 0  # The floor id, which is always 0 since the floor is spawned first.
                elif body_name in self.object_name_to_id.keys():
                    results[-1].obj_id = self.object_name_to_id[body_name]
                else:
                    for obj in self.objects:
                        if body_name in obj.links.keys():
                            results[-1].obj_id = obj.id
                            break
            if return_distance:
                results[-1].distance = ray_result.distance
        return results

    def step(self, func: Optional[Callable[[], None]] = None, step_seconds: Optional[float] = None) -> None:
        """
        Perform a simulation step in the simulator, this is useful when use_static_mode is True.

        :param func: A function to be called after the simulation step.
        :param step_seconds: The number of seconds to step the simulation.
        """
        if self.conf.use_static_mode:
            self.unpause_simulation()
            if func is not None:
                func()
            sleep(self.simulation_time_step if step_seconds is None else step_seconds)
            self.pause_simulation()

    def save_physics_simulator_state(self, state_id: Optional[int] = None, use_same_id: bool = False) -> int:
        if state_id is None:
            self.latest_save_id = 0 if self.latest_save_id is None else self.latest_save_id + int(not use_same_id)
            state_id = self.latest_save_id
        save_name = f"save_{state_id}"
        self.saved_simulator_states[state_id] = self.api_requester.save(save_name)
        return state_id

    def remove_physics_simulator_state(self, state_id: int) -> None:
        save_path = self.saved_simulator_states.pop(state_id)
        shutil.rmtree(Path(save_path).parent)

    def restore_physics_simulator_state(self, state_id: int) -> None:
        self.api_requester.load(self.saved_simulator_states[state_id])

    def set_link_color(self, link: Link, rgba_color: Color):
        logwarn("set_link_color is not implemented in Multiverse")

    def get_link_color(self, link: Link) -> Color:
        logwarn("get_link_color is not implemented in Multiverse")
        return Color()

    def get_colors_of_object_links(self, obj: Object) -> Dict[str, Color]:
        logwarn("get_colors_of_object_links is not implemented in Multiverse")
        return {}

    def set_realtime(self, real_time: bool) -> None:
        logwarn("set_realtime is not implemented as an API in Multiverse, it is configured in the"
                "multiverse configuration file (.muv file) as rtf_required where a value of 1 means real-time")

    def set_gravity(self, gravity_vector: List[float]) -> None:
        logwarn("set_gravity is not implemented in Multiverse")

    def check_object_exists(self, obj: Object) -> bool:
        """
        Check if the object exists in the Multiverse world.

        :param obj: The object.
        :return: True if the object exists, False otherwise.
        """
        return self.api_requester.check_object_exists(obj)



import pathlib
import unittest

from pycram.datastructures.pose import GraspDescription
from pycram.robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    CameraDescription, RobotDescriptionManager
from pycram.datastructures.enums import Arms, GripperState, StaticJointState, Grasp
from pycram.object_descriptors.urdf import ObjectDescription as URDF


class TestRobotDescription(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.path = str(pathlib.Path(__file__).parent.resolve()) + '/../resources/robots/' + "pr2" + '.urdf'
        cls.path_turtlebot = str(pathlib.Path(__file__).parent.resolve()) + '/../resources/robots/' + "turtlebot" + '.urdf'
        cls.urdf_obj = URDF(cls.path)

    def test_robot_description_construct(self):
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        self.assertEqual(robot_description.name, "pr2")
        self.assertEqual(robot_description.base_link, "base_link")
        self.assertEqual(robot_description.torso_link, "torso_lift_link")
        self.assertEqual(robot_description.torso_joint, "torso_lift_joint")
        self.assertTrue(type(robot_description.urdf_object) is URDF)
        self.assertEqual(len(robot_description.links), 88)
        self.assertEqual(len(robot_description.joints), 87)

    def test_kinematic_chain_description_construct(self):
        chain = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        self.assertEqual(chain.name, "left")
        self.assertEqual(chain.start_link, "torso_lift_link")
        self.assertEqual(chain.end_link, "l_wrist_roll_link")
        self.assertEqual(chain.arm_type, Arms.LEFT)
        self.assertTrue("torso_lift_link" in chain.links)
        self.assertTrue("l_wrist_roll_link" in chain.links)
        self.assertEqual(len(chain.links), 10)
        self.assertEqual(len(chain.joints), 7)

    def test_end_effector_description_construct(self):
        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                              self.urdf_obj)
        self.assertEqual(end_effector.name, "left_gripper")
        self.assertEqual(end_effector.start_link, "l_gripper_palm_link")
        self.assertEqual(end_effector.tool_frame, "l_gripper_tool_frame")
        self.assertTrue("l_gripper_palm_link" in end_effector.links)
        self.assertEqual(len(end_effector.links), 11)
        self.assertEqual(len(end_effector.joints), 10)

    def test_camera_description_construct(self):
        camera = CameraDescription("kinect_camera", "wide_stereo_optical_frame", 1.27,
                                   1.60, 0.99483, 0.75049)
        self.assertEqual(camera.name, "kinect_camera")
        self.assertEqual(camera.link_name, "wide_stereo_optical_frame")
        self.assertEqual(camera.minimal_height, 1.27)
        self.assertEqual(camera.maximal_height, 1.60)
        self.assertEqual(camera.horizontal_angle, 0.99483)
        self.assertEqual(camera.vertical_angle, 0.75049)
        self.assertEqual(camera.front_facing_axis, [0, 0, 1])

    def test_kinematic_chain_description_add_to_robot_description(self):
        chain = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        robot_description.add_kinematic_chain_description(chain)
        self.assertTrue(chain in robot_description.kinematic_chains.values())
        self.assertEqual(robot_description.kinematic_chains["left"], chain)
        self.assertEqual(robot_description.get_arm_chain(Arms.LEFT), chain)
        self.assertEqual(len(robot_description.kinematic_chains), 1)

    def test_end_effector_description_add_to_kinematic_chain_description(self):
        chain = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                              self.urdf_obj)
        chain.end_effector = end_effector
        self.assertEqual(chain.end_effector, end_effector)
        self.assertEqual(chain.get_tool_frame(), "l_gripper_tool_frame")

    def test_camera_description_add_to_robot_description(self):
        camera = CameraDescription("kinect_camera", "wide_stereo_optical_frame", 1.27,
                                   1.60, 0.99483, 0.75049)
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        robot_description.add_camera_description(camera)
        self.assertTrue(camera in robot_description.cameras.values())
        self.assertEqual(robot_description.cameras["kinect_camera"], camera)
        self.assertEqual(len(robot_description.cameras), 1)

    def test_kinematic_chain_description_add_static_joint_states(self):
        chain = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        chain.add_static_joint_states(StaticJointState.Park, {'l_shoulder_pan_joint': 1.712,
                                               'l_shoulder_lift_joint': -0.264,
                                               'l_upper_arm_roll_joint': 1.38,
                                               'l_elbow_flex_joint': -2.12,
                                               'l_forearm_roll_joint': 16.996,
                                               'l_wrist_flex_joint': -0.073,
                                               'l_wrist_roll_joint': 0.0})
        self.assertTrue(StaticJointState.Park in chain.static_joint_states)
        self.assertEqual(chain.static_joint_states[StaticJointState.Park], {'l_shoulder_pan_joint': 1.712,
                                                             'l_shoulder_lift_joint': -0.264,
                                                             'l_upper_arm_roll_joint': 1.38,
                                                             'l_elbow_flex_joint': -2.12,
                                                             'l_forearm_roll_joint': 16.996,
                                                             'l_wrist_flex_joint': -0.073,
                                                             'l_wrist_roll_joint': 0.0})

    def test_end_effector_description_add_static_joint_states(self):
        chain = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                              self.urdf_obj)
        chain.end_effector = end_effector
        end_effector.add_static_joint_states(GripperState.OPEN, {'l_gripper_l_finger_joint': 0.548,
                                                                 'l_gripper_r_finger_joint': 0.548})
        self.assertTrue(GripperState.OPEN in end_effector.static_joint_states.keys())
        self.assertEqual(end_effector.static_joint_states[GripperState.OPEN], {'l_gripper_l_finger_joint': 0.548,
                                                                               'l_gripper_r_finger_joint': 0.548})
        self.assertEqual(chain.get_static_gripper_state(GripperState.OPEN), {'l_gripper_l_finger_joint': 0.548,
                                                                             'l_gripper_r_finger_joint': 0.548})

    def test_get_parent_link(self):
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        self.assertEqual("base_link", robot_description.get_parent("torso_lift_joint"))
        self.assertEqual("l_wrist_roll_joint", robot_description.get_parent("l_wrist_roll_link"))

    def test_get_parent_joint(self):
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        self.assertEqual("torso_lift_joint", robot_description.get_parent("torso_lift_link"))
        self.assertEqual("l_wrist_roll_joint", robot_description.get_parent("l_wrist_roll_link"))

    def test_get_child_links(self):
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        self.assertEqual("l_forearm_link", robot_description.get_child("l_forearm_joint", return_multiple_children=True))
        self.assertEqual("sensor_mount_link", robot_description.get_child("sensor_mount_frame_joint"))

    def test_get_child_joints(self):
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        self.assertEqual(["projector_wg6802418_frame_joint", "sensor_mount_frame_joint"],
                         robot_description.get_child("head_plate_frame", return_multiple_children=True))
        self.assertEqual("base_bellow_joint", robot_description.get_child("base_link"))

    def test_get_manipulation_chain(self):
        chain_left = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                              self.urdf_obj)
        chain_right = KinematicChainDescription("right", "torso_lift_link", "r_wrist_roll_link", self.urdf_obj, arm_type=Arms.RIGHT)
        chain_left.end_effector = end_effector
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        robot_description.add_kinematic_chain_description(chain_left)
        robot_description.add_kinematic_chain_description(chain_right)
        self.assertEqual(chain_left, robot_description.kinematic_chains["left"])
        self.assertEqual(chain_left, robot_description.get_arm_chain(Arms.LEFT))
        self.assertEqual([chain_left], robot_description.get_manipulator_chains())
        self.assertTrue(chain_right not in robot_description.get_manipulator_chains())

    def test_get_arm_chains(self):
        chain_left = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        chain_right = KinematicChainDescription("right", "torso_lift_link", "r_wrist_roll_link", self.urdf_obj, arm_type=Arms.RIGHT)
        robot_description = RobotDescription("pr2", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        robot_description.add_kinematic_chain_description(chain_left)
        robot_description.add_kinematic_chain_description(chain_right)
        self.assertEqual(chain_left, robot_description.get_arm_chain(Arms.LEFT))
        self.assertEqual([chain_left, chain_right], robot_description.get_arm_chain(Arms.BOTH))

    def test_get_tool_frame(self):
        chain_left = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link", self.urdf_obj,
                                          arm_type=Arms.LEFT)
        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                                self.urdf_obj)
        chain_left.end_effector = end_effector
        self.assertEqual("l_gripper_tool_frame", chain_left.get_tool_frame())

    def test_robot_description_manager_singelton(self):
        rdm = RobotDescriptionManager()
        rdm2 = RobotDescriptionManager()
        self.assertEqual(rdm, rdm2)
        self.assertIs(rdm, rdm2)

    def test_register_robot_description(self):
        robot_description = RobotDescription("pr2_test1", "base_link", "torso_lift_link", "torso_lift_joint", self.path)
        rdm = RobotDescriptionManager()
        rdm.register_description(robot_description)
        self.assertEqual(rdm.descriptions["pr2_test1"], robot_description)
        self.assertTrue(robot_description in rdm.descriptions.values())

    def test_load_robot_description(self):
        rdm = RobotDescriptionManager()
        rdm.load_description(self.urdf_obj.name) # loads description by the name of the urdf
        self.assertIs(RobotDescription.current_robot_description, rdm.descriptions["pr2"])

    def test_robot_description_turtlebot(self):
        robot_description = RobotDescription("turtlebot", "base_link", "base_link", "base_joint", self.path_turtlebot)
        self.assertEqual(robot_description.name, "turtlebot")
        self.assertEqual(robot_description.base_link, "base_link")
        self.assertEqual(robot_description.torso_link, "base_link")
        self.assertEqual(robot_description.torso_joint, "base_joint")
        self.assertTrue(type(robot_description.urdf_object) is URDF)
        self.assertEqual(len(robot_description.links), 11)
        self.assertEqual(len(robot_description.joints), 10)

    def test_grasp_descriptions(self):
        grasp1 = GraspDescription(Grasp.LEFT, Grasp.TOP, True)
        grasp2 = GraspDescription(Grasp.BACK, Grasp.BOTTOM, False)
        grasp3 = GraspDescription(Grasp.RIGHT, None, True)

        grasp1_quat = [0.7071067811865476, 0.0, -0.7071067811865476, 0.0]
        grasp2_quat = [0.7071067811865476, 0.0, 0.7071067811865476, 0.0]
        grasp3_quat = [0.5, 0.5, 0.5, 0.5]

        end_effector = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                              self.urdf_obj)

        end_effector.update_all_grasp_orientations([0, 0, 0, 1])

        self.assertEqual(grasp1_quat, end_effector.grasps[grasp1])
        self.assertEqual(grasp2_quat, end_effector.grasps[grasp2])
        self.assertEqual(grasp3_quat, end_effector.grasps[grasp3])




from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, TorsoState, StaticJointState
from ..ros import get_ros_package_path

filename = get_ros_package_path('pycram') + '/resources/robots/' + "rollin_justin" + '.urdf'

justin_description = RobotDescription("rollin_justin", "base_link", "torso2", "torso2_joint",
                                      filename)

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left", "base_link", "left_arm7",
                                     justin_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {"torso1_joint": 0,
                                          "torso2_joint": 0,
                                          "torso3_joint": 0.174533,
                                          "torso4_joint": 0,
                                          "left_arm1_joint": 0,
                                          "left_arm2_joint": -1.9,
                                          "left_arm3_joint": 0,
                                          "left_arm4_joint": 1,
                                          "left_arm5_joint": 0,
                                          "left_arm6_joint": -1,
                                          "left_arm7_joint": 0})

justin_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "left_arm7", "l_gripper_tool_frame",
                                      justin_description.urdf_object)
left_gripper.add_static_joint_states(GripperState.OPEN, {"left_1thumb_base_joint": 0.0,
                                                         "left_1thumb1_joint": 0.0,
                                                         "left_1thumb2_joint": 0.0,
                                                         "left_1thumb3_joint": 0.0,
                                                         "left_1thumb4_joint": 0.0,
                                                         "left_2tip_base_joint": 0.0,
                                                         "left_2tip1_joint": 0.0,
                                                         "left_2tip2_joint": 0.0,
                                                         "left_2tip3_joint": 0.0,
                                                         "left_2tip4_joint": 0.0,
                                                         "left_3middle_base_joint": 0.0,
                                                         "left_3middle1_joint": 0.0,
                                                         "left_3middle2_joint": 0.0,
                                                         "left_3middle3_joint": 0.0,
                                                         "left_3middle4_joint": 0.0,
                                                         "left_4ring_base_joint": 0.0,
                                                         "left_4ring1_joint": 0.0,
                                                         "left_4ring2_joint": 0.0,
                                                         "left_4ring3_joint": 0.0,
                                                         "left_4ring4_joint": 0.0})
left_gripper.add_static_joint_states(GripperState.CLOSE, {"left_1thumb_base_joint": 0.0,
                                                          "left_1thumb1_joint": 0.523599,
                                                          "left_1thumb2_joint": 1.50098,
                                                          "left_1thumb3_joint": 1.76278,
                                                          "left_1thumb4_joint": 1.76278,
                                                          "left_2tip_base_joint": 0.0,
                                                          "left_2tip1_joint": 0.523599,
                                                          "left_2tip2_joint": 1.50098,
                                                          "left_2tip3_joint": 1.76278,
                                                          "left_2tip4_joint": 1.76278,
                                                          "left_3middle_base_joint": 0.0,
                                                          "left_3middle1_joint": 0.523599,
                                                          "left_3middle2_joint": 1.50098,
                                                          "left_3middle3_joint": 1.76278,
                                                          "left_3middle4_joint": 1.76278,
                                                          "left_4ring_base_joint": 0.0,
                                                          "left_4ring1_joint": 0.523599,
                                                          "left_4ring2_joint": 1.50098,
                                                          "left_4ring3_joint": 1.76278,
                                                          "left_4ring4_joint": 1.76278})

left_gripper.end_effector_type = GripperType.FINGER
# left_gripper.opening_distance = 0.548
left_arm.end_effector = left_gripper

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right", "base_link", "right_arm7",
                                      justin_description.urdf_object, arm_type=Arms.RIGHT)

right_arm.add_static_joint_states(StaticJointState.Park, {"torso1_joint": 0,
                                           "torso2_joint": 0,
                                           "torso3_joint": 0.174533,
                                           "torso4_joint": 0,
                                           "right_arm1_joint": 0,
                                           "right_arm2_joint": -1.9,
                                           "right_arm3_joint": 0,
                                           "right_arm4_joint": 1,
                                           "right_arm5_joint": 0,
                                           "right_arm6_joint": -1,
                                           "right_arm7_joint": 0})

justin_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################
right_gripper = EndEffectorDescription("right_gripper", "right_arm7", "r_gripper_tool_frame",
                                       justin_description.urdf_object)
right_gripper.add_static_joint_states(GripperState.OPEN, {"right_1thumb_base_joint": 0.0,
                                                          "right_1thumb1_joint": 0.0,
                                                          "right_1thumb2_joint": 0.0,
                                                          "right_1thumb3_joint": 0.0,
                                                          "right_1thumb4_joint": 0.0,
                                                          "right_2tip_base_joint": 0.0,
                                                          "right_2tip1_joint": 0.0,
                                                          "right_2tip2_joint": 0.0,
                                                          "right_2tip3_joint": 0.0,
                                                          "right_2tip4_joint": 0.0,
                                                          "right_3middle_base_joint": 0.0,
                                                          "right_3middle1_joint": 0.0,
                                                          "right_3middle2_joint": 0.0,
                                                          "right_3middle3_joint": 0.0,
                                                          "right_3middle4_joint": 0.0,
                                                          "right_4ring_base_joint": 0.0,
                                                          "right_4ring1_joint": 0.0,
                                                          "right_4ring2_joint": 0.0,
                                                          "right_4ring3_joint": 0.0,
                                                          "right_4ring4_joint": 0.0})
right_gripper.add_static_joint_states(GripperState.CLOSE, {"right_1thumb_base_joint": 0.0,
                                                           "right_1thumb1_joint": 0.523599,
                                                           "right_1thumb2_joint": 1.50098,
                                                           "right_1thumb3_joint": 1.76278,
                                                           "right_1thumb4_joint": 1.76278,
                                                           "right_2tip_base_joint": 0.0,
                                                           "right_2tip1_joint": 0.523599,
                                                           "right_2tip2_joint": 1.50098,
                                                           "right_2tip3_joint": 1.76278,
                                                           "right_2tip4_joint": 1.76278,
                                                           "right_3middle_base_joint": 0.0,
                                                           "right_3middle1_joint": 0.523599,
                                                           "right_3middle2_joint": 1.50098,
                                                           "right_3middle3_joint": 1.76278,
                                                           "right_3middle4_joint": 1.76278,
                                                           "right_4ring_base_joint": 0.0,
                                                           "right_4ring1_joint": 0.523599,
                                                           "right_4ring2_joint": 1.50098,
                                                           "right_4ring3_joint": 1.76278,
                                                           "right_4ring4_joint": 1.76278})

right_gripper.end_effector_type = GripperType.FINGER
# right_gripper.opening_distance = 0.548
right_arm.end_effector = right_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "torso1", "torso4",
                                  justin_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso2_joint": 0,
                                                "torso3_joint": 0.174533,
                                                "torso4_joint": 0})

torso.add_static_joint_states(TorsoState.MID, {"torso2_joint": -0.8,
                                               "torso3_joint": 1.57,
                                               "torso4_joint": -0.77})

torso.add_static_joint_states(TorsoState.LOW, {"torso2_joint": -0.9,
                                               "torso3_joint": 2.33874,
                                               "torso4_joint": -1.57})

justin_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
# real camera unknown at the moment of writing (also missing in urdf), so using dummy camera for now
camera = CameraDescription("dummy_camera", "head2", 1.27,
                           1.85, 0.99483, 0.75049,
                           [1, 0, 0])
justin_description.add_camera_description(camera)

################################## Neck ##################################
justin_description.add_kinematic_chain("neck", "torso4", "head2")
justin_description.set_neck("head1_joint", "head2_joint")

################################# Grasps ##################################
orientation = [0.707, -0.707, 0.707, -0.707]
left_gripper.update_all_grasp_orientations(orientation)

orientation = [0.707, 0.707, 0.707, 0.707]
right_gripper.update_all_grasp_orientations(orientation)


################################# Additionals ##################################
# justin_description.set_costmap_offset(0.5)
# justin_description.set_palm_axis([0, 0, 1])

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(justin_description)



import logging

from .datastructures.enums import TaskStatus
from .designator import DesignatorDescription, BaseMotion
from .failures import PlanFailure
from threading import Lock
from typing_extensions import Union, Tuple, Any, List, Optional, Type, Callable
from .language import LanguageMixin, MonitorNode, MonitorPlan
from .plan import Plan
from .process_module import ProcessModule


class FailureHandling(LanguageMixin):
    """
    Base class for failure handling mechanisms in automated systems or workflows.

    This class provides a structure for implementing different strategies to handle
    failures that may occur during the execution of a plan or process. It is designed
    to be extended by subclasses that implement specific failure handling behaviors.
    """

    def __init__(self, plan: Optional[Plan] = None):
        """
        Initializes a new instance of the FailureHandling class.

        :param Union[DesignatorDescription, MonitorNode] plan: The description or context of the task
        or process for which the failure handling is being set up.
        """
        self.plan = plan

    def perform(self):
        """
        Abstract method to perform the failure handling mechanism.

        This method should be overridden in subclasses to implement the specific
        behavior for handling failures.

        :raises NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError()


class Retry(FailureHandling):
    """
    A subclass of FailureHandling that implements a retry mechanism.

    This class represents a specific failure handling strategy where the system
    attempts to retry a failed action a certain number of times before giving up.
    """
    max_tries: int
    """
    The maximum number of attempts to retry the action.
    """

    def __init__(self, plan: MonitorPlan, max_tries: int = 3):
        """
        Initializes a new instance of the Retry class.

        :param plan: The description or context of the task or process for which the retry mechanism is being set up.
        :param max_tries: The maximum number of attempts to retry. Defaults to 3.
        """
        super().__init__(plan)
        self.max_tries = max_tries

    def perform(self) -> List[Any]:
        """
        Implementation of the retry mechanism.

        This method attempts to perform the action specified in the designator_description.
        If the action fails, it is retried up to max_tries times. If all attempts fail,
        the last exception is raised.

        :raises PlanFailure: If all retry attempts fail.
        """
        tries = 0
        for action in iter(self.plan):
            tries += 1
            try:
                action.perform()
                break
            except PlanFailure as e:
                if tries >= self.max_tries:
                    raise e


class RetryMonitor(FailureHandling):
    """
    A subclass of FailureHandling that implements a retry mechanism that works with a Monitor.

    This class represents a specific failure handling strategy that allows us to retry a demo that is
    being monitored, in case that monitoring condition is triggered.
    """
    max_tries: int
    """
    The maximum number of attempts to retry the action.
    """
    recovery: dict
    """
    A dictionary that maps exception types to recovery actions
    """

    def __init__(self, monitor_plan: Plan, max_tries: int = 3, recovery: dict = None):
        """
        Initializes a new instance of the RetryMonitor class.

        :param MonitorNode monitor_plan: The Monitor instance to be used.
        :param int max_tries: The maximum number of attempts to retry. Defaults to 3.
        :param dict recovery: A dictionary that maps exception types to recovery actions. Defaults to None.
        """
        super().__init__(monitor_plan)
        self.max_tries = max_tries
        self.lock = Lock()
        if recovery is None:
            self.recovery = {}
        else:
            if not isinstance(recovery, dict):
                raise ValueError(
                    "Recovery must be a dictionary with exception types as keys and Language instances as values.")
            for key, value in recovery.items():
                if not issubclass(key, BaseException):
                    raise TypeError("Keys in the recovery dictionary must be exception types.")
                if not callable(value):
                    raise TypeError("Values in the recovery dictionary must be instances of the Language class.")
            self.recovery = recovery

    def perform(self) -> List[Any]:
        """
        This method attempts to perform the Monitor + plan specified in the designator_description. If the action
        fails, it is retried up to max_tries times. If all attempts fail, the last exception is raised. In every
        loop, we need to clear the kill_event, and set all relevant 'interrupted' variables to False, to make sure
        the Monitor and plan are executed properly again.

        :raises PlanFailure: If all retry attempts fail.

        :return: The state of the execution performed, as well as a flattened list of the
        results, in the correct order
        """

        def reset_interrupted(child):
            child.status = TaskStatus.CREATED
            try:
                for sub_child in child.children:
                    reset_interrupted(sub_child)
            except AttributeError:
                pass

        def flatten(result):
            flattened_list = []
            if result:
                for item in result:
                    if isinstance(item, list):
                        flattened_list.extend(item)
                    else:
                        flattened_list.append(item)
                return flattened_list
            return None

        status, res = None, None
        with self.lock:
            tries = 0
            while True:
                self.plan.interrupted = False
                for child in self.plan.root.children:
                    reset_interrupted(child)
                try:
                    if tries >= 1:
                        self.plan.re_perform()
                        break
                    else:
                        res = self.plan.perform()
                        break
                except PlanFailure as e:
                    tries += 1
                    if tries >= self.max_tries:
                        raise e
                    exception_type = type(e)
                    if exception_type in self.recovery:
                        self.recovery[exception_type]()
        return flatten(res)


def try_action(action: Any, failure_type: Type[Exception], max_tries: int = 3):
    """
    A generic function to retry an action a certain number of times before giving up, with a specific failure type.

    :param action: The action to be performed, it must have a perform() method.
    :param failure_type: The type of exception to catch.
    :param max_tries: The maximum number of attempts to retry the action. Defaults to 3.
    """
    return try_method(method=lambda: action.perform(),
                      failure_type=failure_type,
                      max_tries=max_tries,
                      name="action")


def try_motion(motion: ProcessModule, motion_designator_instance: BaseMotion,
               failure_type: Type[Exception], max_tries: int = 3):
    """
    A generic function to retry a motion a certain number of times before giving up, with a specific exception.

    :param motion: The motion to be executed.
    :param motion_designator_instance: The instance of the motion designator that has the description of the motion.
    :param failure_type: The type of exception to catch.
    :param max_tries: The maximum number of attempts to retry the motion.
    """
    return try_method(method=lambda: motion.execute(motion_designator_instance),
                      failure_type=failure_type,
                      max_tries=max_tries,
                      name="motion")


def try_method(method: Callable, failure_type: Type[Exception], max_tries: int = 3, name: str = "method"):
    """
    A generic function to retry a method a certain number of times before giving up, with a specific exception.

    :param method: The method to be called.
    :param failure_type: The type of exception to catch.
    :param max_tries: The maximum number
    :param name: The name of the method to be called.
    """
    current_retry = 0
    result = None
    while current_retry < max_tries:
        try:
            result = method()
            break
        except failure_type as e:
            logging.debug(f"Caught exception {e} during {name} execution {method}. Retrying...")
            current_retry += 1
    if current_retry == max_tries:
        logging.error(f"Failed to execute {name} {method} after {max_tries} retries.")
    return result



import atexit

import time

from geometry_msgs.msg import TransformStamped
from sensor_msgs.msg import JointState
from ..datastructures.world import World
from ..robot_description import RobotDescription
from ..datastructures.pose import PoseStamped
from ..ros import  Time, Duration, sleep
from ..ros import  wait_for_message, create_timer


class RobotStateUpdater:
    """
    Updates the robot in the World with information of the real robot published to ROS topics.
    Infos used to update the robot are:

        * The current pose of the robot
        * The current joint state of the robot
    .. Note:: This class can only be used if the topics are present in the RSO network on the real robot/world,
    hence it is not testable in the CI.
    """

    def __init__(self, tf_topic: str, joint_state_topic: str):
        """
        The robot state updater uses a TF topic and a joint state topic to get the current state of the robot.

        :param tf_topic: Name of the TF topic, needs to publish geometry_msgs/TransformStamped
        :param joint_state_topic: Name of the joint state topic, needs to publish sensor_msgs/JointState
        """
        self.tf_listener = tf.TransformListener()
        self.tf_listener.clear()

        sleep(1)
        self.tf_topic = tf_topic
        self.joint_state_topic = joint_state_topic
        self.tf_timer = create_timer(Duration().from_sec(0.1), self._subscribe_tf)
        self.joint_state_timer = create_timer(Duration().from_sec(0.1), self._subscribe_joint_state)

        atexit.register(self._stop_subscription)

    def _subscribe_tf(self, msg: TransformStamped) -> None:
        """
        Callback for the TF timer, will do a lookup of the transform between map frame and the robot base frame.

        :param msg: TransformStamped message published to the topic
        """
        self.tf_listener.waitForTransform("map", RobotDescription.current_robot_description.base_link, Time(0.0), Duration(5))
        trans, rot = self.tf_listener.lookupTransform("map", RobotDescription.current_robot_description.base_link, Time(0.0))
        World.robot.set_pose(PoseStamped(trans, rot))

    def _subscribe_joint_state(self, msg: JointState) -> None:
        """
        Sets the current joint configuration of the robot in the world to the configuration published on the
        topic. Since this uses rospy.wait_for_message which can have errors when used with threads there might be an
        attribute error in the rospy implementation.

        :param msg: JointState message published to the topic.
        """
        try:
            msg = wait_for_message(self.joint_state_topic, JointState)
            for name, position in zip(msg.name, msg.position):
                World.robot.set_joint_position(name, position)
        except AttributeError:
            pass

    def _stop_subscription(self) -> None:
        """
        Stops the Timer for TF and joint states and therefore the updating of the robot in the world.
        """
        self.tf_timer.shutdown()
        self.joint_state_timer.shutdown()


class EnvironmentStateUpdater:
    """
    Updates the environment in the World with information of the real environment published to ROS topics.
    Infos used to update the envi are:
        * The current pose of the environment
        * The current joint state of the environment

    .. Note:: This class can only be used if the topics are present in the RSO network on the real robot/world,
    hence it is not testable in the CI.
    """

    def __init__(self, tf_topic: str, joint_state_topic: str):
        """
        The environment state updater uses a TF topic and a joint state topic to get the current state of the environment.

        :param tf_topic: Name of the TF topic, needs to publish geometry_msgs/TransformStamped
        :param joint_state_topic: Name of the joint state topic, needs to publish sensor_msgs/JointState
        """
        self.tf_listener = tf.TransformListener()
        time.sleep(1)
        self.tf_topic = tf_topic
        self.joint_state_topic = joint_state_topic

        self.joint_state_timer = create_timer(Duration(0.1), self._subscribe_joint_state)

        atexit.register(self._stop_subscription)

    def _subscribe_joint_state(self, msg: JointState) -> None:
        """
        Sets the current joint configuration of the environment in the world to the configuration published on the topic.
        Since this uses rospy.wait_for_message which can have errors when used with threads there might be an attribute error
        in the rospy implementation.

        :param msg: JointState message published to the topic.
        """
        try:
            msg = wait_for_message(self.joint_state_topic, JointState)
            for name, position in zip(msg.name, msg.position):
                try:
                    # Attempt to get the joint state. This might throw a KeyError if the joint name doesn't exist
                    if World.environment.get_joint_state(name) is None:
                        continue
                    # Set the joint state if the joint exists
                    World.environment.set_joint_state(name, position)
                except KeyError:
                    # Handle the case where the joint name does not exist
                    pass
        except AttributeError:
            pass

    def _stop_subscription(self) -> None:
        """
        Stops the Timer for TF and joint states and therefore the updating of the environment in the world.
        """
        self.joint_state_timer.shutdown()
        self.joint_state_timer.shutdown()



def create_action_client(topic_name: str, action_message) -> None:
    """
    Create an action client for the given topic name and action message type.

    Args:
        topic_name (str): The name of the action topic.
        action_message: The action message type.

    Returns:
        None
    """
    raise NotImplementedError("No ROS environment detected.")


import sys

from ..ros import  create_action_client
from ..ros import  logwarn, loginfo
from ..ros import  get_node_names

from typing import Callable


# Global variables for shared resources
nav_action_client = None
is_init = False


def create_nav_action_client():
    """Creates a new action client for the move_base interface."""
    client = create_action_client("move_base", MoveBaseAction)
    loginfo("Waiting for move_base action server")
    client.wait_for_server()
    return client


def init_nav_interface(func: Callable) -> Callable:
    """Ensures initialization of the navigation interface before function execution."""

    def wrapper(*args, **kwargs):
        from move_base_msgs.msg import MoveBaseAction, MoveBaseGoal
        from geometry_msgs.msg import PoseStamped


        global is_init
        global nav_action_client

        if is_init:
            return func(*args, **kwargs)

        if "move_base_msgs" not in sys.modules:
            logwarn("Could not initialize the navigation interface: move_base_msgs not imported")
            return

        if "/move_base" in get_node_names():
            nav_action_client = create_nav_action_client()
            loginfo("Successfully initialized navigation interface")
            is_init = True
        else:
            logwarn("Move_base is not running, could not initialize navigation interface")
            return

        return func(*args, **kwargs)

    return wrapper


@init_nav_interface
def query_pose_nav(navpose: 'PoseStamped'):
    """Sends a goal to the move_base service, initiating robot navigation to a given pose."""
    global nav_action_client
    global query_result

    def active_callback():
        loginfo("Sent query to move_base")

    def done_callback(state, result):
        loginfo("Finished moving")
        global query_result
        query_result = result

    goal_msg = MoveBaseGoal()
    goal_msg.target_pose = navpose
    nav_action_client.send_goal(goal_msg, active_cb=active_callback, done_cb=done_callback)

    nav_action_client.wait_for_result()

    return query_result


def cancel_nav():
    """Cancels the current navigation goal."""
    global nav_action_client
    nav_action_client.cancel_all_goals()



import fnmatch
import os

import requests
from bs4 import BeautifulSoup
from ..ros import  ros_tools
from typing_extensions import Dict, Set
from ..ros import  logging as log

filename = ros_tools.get_ros_package_path('pycram')


class ProcThorInterface:
    def __init__(self, base_url="http://procthor.informatik.uni-bremen.de:5000",
                 source_folder=os.path.join(ros_tools.get_ros_package_path('pycram'), "resources/tmp/")):
        self.base_url = base_url
        self.source_folder = source_folder
        self.stored_environments = []
        self.stored_environments.extend(self.get_all_environments_stored_below_directory(self.source_folder))

    def _download_file(self, base_url: str, full_url: str, folder: str) -> str:
        """
        Downloads the file given in full_url and stores it into folder. If necessary it will create the same folder
        structure. For this purpose the base_url is necessary to decide what is folder structure and what is just url.


        :param base_url: Base url as string from
        :param full_url: Full url of the file to be downloaded
        :param folder: Folder where the file should be stored
        :return: The local file name of the downloaded file
        """
        tree_structure = full_url.replace(base_url, '')
        if tree_structure.startswith("/"):
            tree_structure = tree_structure[1:]
        full_storage_path = os.path.join(folder, tree_structure)
        if not os.path.exists(full_storage_path[:full_storage_path.rfind('/') + 1]):
            os.makedirs(full_storage_path[:full_storage_path.rfind('/') + 1], exist_ok=True)
        local_filename = os.path.join(folder, full_url.split('/')[-1])
        # Send HTTP GET request to fetch the file
        with requests.get(full_url, stream=True) as r:
            r.raise_for_status()
            # Write the content to a file
            with open(full_storage_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return local_filename

    # Function to get the list of files in a directory on the web server
    def _get_files_list(self, base_url: str) -> Set[str]:
        """
        Function to get the list of files in a directory on the web server.

        :param base_url: Base url as string from
        :return: Set of all files found below the given url
        """
        # Send GET request to the URL
        response = requests.get(base_url)
        # Parse the HTML response
        soup = BeautifulSoup(response.text, 'html.parser')
        # Find all links on the page
        links = soup.find_all('a')
        file_list = set()
        files_urls = set()
        for link in links:
            if link.get('href').endswith('/') and not link.get('href').startswith('/'):
                files_urls = files_urls.union(self._get_files_list(os.path.join(base_url, link['href'])))

            elif link.get('href').endswith(('.usda', '.urdf', '.stl', '.usd', '.hdr')):
                files_urls.add(os.path.join(base_url, link['href']))
            else:
                log.logwarn("Ignored File: {}".format(link.get('href')))
        return files_urls

    # Main function to download all files from a directory
    def download_all_files_from_URL(self, base_url: str, folder: str) -> None:
        """
        Main function to download all files from a given path. Files will be stored in folder.

        :param base_url: Base url as string from
        :param folder: folder where the files should be stored
        :return: None
        """
        # Ensure the folder exists
        os.makedirs(folder, exist_ok=True)
        # Get the list of files
        files = self._get_files_list(base_url)
        # create_folder_structure(base_url,folder,files)
        # Download each file
        for file_url in files:
            log.loginfo(f"Downloading {file_url}...")
            self._download_file(base_url, file_url, folder)
        log.loginfo("All files downloaded.")
        return None

    # Returns a list of all the urdf files in a given folder structure if non is given uses the base_source_folder
    # only look for urdf, but can be easly adapted
    def get_all_environments_stored_below_directory(self, source_folder: str) -> list:
        """
        Returns a list of dictionaries that contains the name and the storage_place of all urdf files below a given
        folder. Only looks for urdf.

        :param source_folder: Base path as string from
        :return: map from name to storage place
        """
        urdf_files = []
        urdf_file = {}
        for root, dirs, files in os.walk(source_folder):
            for filename in fnmatch.filter(files, '*.urdf'):
                urdf_file["name"] = filename.rsplit('.', 1)[0]
                urdf_file["storage_place"] = os.path.join(root, filename)
                urdf_files.append(urdf_file.copy())
        return urdf_files

    def download_one_random_environment(self) -> Dict[str, str]:
        """
        Downloads one random environment currently not present.

        :param: None
        :return: Dictionary of name and storage place
        """
        endpoint = "GetRandomEnvironment"
        full_url = os.path.join(self.base_url, endpoint)
        response = requests.get(full_url)
        self.download_all_files_from_URL(response.json()["storage_place"], self.source_folder)
        self.stored_environments.append(response.json())
        return response.json()

    def download_num_random_environment(self, num_of_environments: int) -> list:
        """
        Downloads given amount of  random environment currently not present.

        :param num_of_environments: Amount of environments that get downloaded
        :return: List of Dictionaries of name and storage place
        """
        endpoint = "GetNumEnvironment"
        full_url = os.path.join(self.base_url, endpoint, str(num_of_environments))
        response = requests.get(full_url)
        for env in response.json():
            self.download_all_files_from_URL(env["storage_place"], self.source_folder)
            self.stored_environments.append(env)
        return response.json()
